import requests
from objict import objict


def getURL(host, path):
    if host.endswith("/"):
        host = host[:-1]
    if path.startswith("/"):
        path = path[1:]
    if host.startswith("http"):
        return "{host}/{path}"
    return f"https://{host}/{path}" 


def REQUEST(method, host, path, data=None, params=None, headers=None, files=None, session=None, post_json=False):
    url = getURL(host, path)
    if headers is None:
        headers = objict()

    fields = dict(url=url, params=params, files=files)
    if isinstance(data, dict):
        fields["data"] = objict.fromdict(data).toJSON()
    elif data is not None:
        fields["data"] = data

    # headers["Accept"] = 'application/json'
    if post_json:
        headers['Content-type'] = 'application/json'
    fields["headers"] = headers
    if session is None:
        session = requests
    try:
        res = getattr(session, method.lower())(**fields)
    except Exception as err:
        return objict(status=False, status_code=500, error=str(err))
    return processResponse(res, url)


def processResponse(res, url):
    data = res.text
    try:
        data = objict.fromdict(res.json())
    except Exception as err:
        return objict(status=False, status_code=res.status_code, data=res.text, error=str(err))

    if res.status_code not in [200, 201]:
        if isinstance(data, dict) and data.error:
            data.status = False
            data.status_code = res.status_code
            return data
        return objict(status=False, status_code=res.status_code, error=res.text)
    if data.data:
        data = data.data
    return objict(status=True, status_code=res.status_code, data=data)



import pyqrcode
import io
import base64

def generateQRCode(data, error="H", version=None, mode=None, img_format="png", scale=1):
    # L, M, Q, or H; each level can correct up to 7, 15, 25, or 30
    c = pyqrcode.create(data, error=error)
    s = io.BytesIO()
    if img_format == "base64":
        c.png(s,scale=scale)
        return base64.b64encode(s.getvalue()).decode("ascii")
    elif img_format == "svg":
        c.svg(s, scale=scale) # module_color="#000000"
        return s.getvalue()
    c.png(s,scale=scale)
    return s.getvalue()



from django.db import models
from django.conf import settings

from django.template.loader import render_to_string
from rest.models import RestModel
from account.models import Group, Member
from . import phone_util


class SMS(models.Model, RestModel):
    class RestMeta:
        CAN_CREATE = False
        DEFAULT_SORT = "-created"
        QUERY_FIELDS = ["endpoint", "srcpoint"]
        SEARCH_FIELDS = ["endpoint", "srcpoint"]
        SEARCH_TERMS = [
            ("phone", "endpoint"),
            ("to", "endpoint"),
            ("from", "srcpoint")]
        FILTER_FIELDS = ["state", "created", "owner", "group"]
        GRAPHS = {
            "basic": {
                "graphs": {
                    "by": "basic",
                    "to": "basic"
                }
            },

            "default": {
                "graphs": {
                    "self": "basic",
                }
            },

            "detailed": {
                "extra": ["number_info"],
                "graphs": {
                    "by": "basic",
                    "to": "basic"
                }
            }
        }

    sid = models.CharField(
        max_length=125, blank=True, null=True,
        default=None, db_index=True)
    created = models.DateTimeField(auto_now_add=True, editable=False)
    group = models.ForeignKey(
        Group, related_name="telephony_sms",
        blank=True, null=True, default=None,
        on_delete=models.CASCADE)
    by = models.ForeignKey(
        Member, related_name="telephony_sms_out",
        blank=True, null=True, default=None,
        on_delete=models.CASCADE)
    to = models.ForeignKey(
        Member, related_name="telephony_sms_in",
        blank=True, null=True, default=None,
        on_delete=models.CASCADE)
    # the mechanism , sms, push, etc
    transport = models.CharField(max_length=125, blank=True, null=True, default=None)
    # this can be email, phone, etc
    endpoint = models.CharField(
        max_length=200, blank=True,
        null=True, default=None, db_index=True)
    # this can be email, phone, etc
    srcpoint = models.CharField(max_length=200, blank=True, null=True, default=None)

    # direction of message
    is_inbound = models.BooleanField(default=False)

    status = models.CharField(max_length=64, default=None, null=True, db_index=True)
    reason = models.TextField(default=None, null=True)

    message = models.TextField()

    @property
    def number_info(self):
        if self.is_inbound:
            return PhonenumberInfo.lookup(self.srcpoint)
        return PhonenumberInfo.lookup(self.endpoint)

    @staticmethod
    def log_incoming(request):
        """
        {
            "AccountSid": "ACb27c3a393ddca77bc32f0757c98b1a59",
            "ApiVersion": "2010-04-01",
            "Body": "Testing",
            "From": "+19496065381",
            "FromCity": "LAGUNA NIGUEL",
            "FromCountry": "US",
            "FromState": "CA",
            "FromZip": "92692",
            "MessageSid": "SM53f8730fa9bf22ccaf51d3ec8bd87fa9",
            "NumMedia": "0",
            "NumSegments": "1",
            "SmsMessageSid": "SM53f8730fa9bf22ccaf51d3ec8bd87fa9",
            "SmsSid": "SM53f8730fa9bf22ccaf51d3ec8bd87fa9",
            "SmsStatus": "received",
            "To": "+19493157900",
            "ToCity": "LAGUNA BEACH",
            "ToCountry": "US",
            "ToState": "CA",
            "ToZip": "92656"
        }
        """
        msg = SMS(
            endpoint=request.DATA.get("To"),
            message=request.DATA.get("Body"),
            srcpoint=request.DATA.get("From"),
            transport="sms", is_inbound=True)
        msg.sid = request.DATA.get("MessageSid")
        msg.status = request.DATA.get("MessageStatus")
        msg.save()
        return None

    @staticmethod
    def normalizePhone(phonenum):
        return phone_util.normalize(phonenum)

    @staticmethod
    def broadcast(to, message, by=None, group=None, transport="sms", srcpoint=None):
        if srcpoint is None:
            srcpoint = settings.TELEPHONY_DEFAULT_SRC

        srcpoint = phone_util.normalize(srcpoint)
        msg = None
        for t in to:
            to_num = t.getProperty("phone")
            if to_num and len(to_num) > 4:
                SMS.send(to_num, message, by=by, group=group, transport=transport, srcpoint=srcpoint)
        return msg

    @staticmethod
    def sendTemplate(endpoint, template, context, by=None, group=None, to=None, transport="sms", srcpoint=None):
        message = render_to_string(template, context)
        return SMS.send(endpoint, message, by, group, to, transport, srcpoint)

    @staticmethod
    def send(endpoint, message, by=None, group=None, to=None, transport="sms", srcpoint=None, verify=True):
        if srcpoint is None:
            srcpoint = settings.TELEPHONY_DEFAULT_SRC

        # format the to srcpoint
        if endpoint.startswith("555") or endpoint.startswith("+1 555"):
            # HACK: send sms to ians work phone
            endpoint = settings.TELEPHONY_555_TO

        endpoint = phone_util.normalize(endpoint)
        srcpoint = phone_util.normalize(srcpoint)
        if verify:
            to_info = PhonenumberInfo.lookup(endpoint)
            if not to_info.sms_allowed:
                msg = SMS(endpoint=endpoint, message=message, srcpoint=srcpoint, transport=transport, by=by, to=to, group=group)
                msg.status = "send failed"
                msg.reason = "number valid: {} number kind: {} cannot receive sms".format(to_info.is_valid, to_info.kind)
                msg.save()
                return msg
        resp = phone_util.sendSMS(endpoint, srcpoint, message)
        msg = SMS(endpoint=endpoint, message=message, srcpoint=srcpoint, transport=transport, by=by, to=to, group=group)
        msg.sid = resp.sid
        msg.status = resp.status
        msg.reason = resp.error_message
        msg.save()
        return msg


class PhonenumberInfo(models.Model, RestModel):
    class RestMeta:
        CAN_CREATE = False
        QUERY_FIELDS = ["number", "owner_name"]
        SEARCH_FIELDS = ["number", "owner_name"]
        SEARCH_TERMS = [
            ("phone", "number"),
            ("name", "owner_name"),
            ("kind", "kind")]

    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)
    number = models.CharField(max_length=200, db_index=True)
    owner_name = models.CharField(max_length=200, default=None, null=True)
    owner_kind = models.CharField(max_length=200, default=None, null=True)
    country = models.CharField(max_length=64, default=None, null=True)
    carrier_name = models.CharField(max_length=64, default=None, null=True)
    kind = models.CharField(max_length=64, default=None, null=True)
    is_valid = models.BooleanField(default=False)
    is_verified = models.BooleanField(default=False)
    sms_allowed = models.BooleanField(default=False)

    def refresh(self):
        resp = phone_util.lookup(self.number)
        if resp.success:
            self.owner_name = resp.owner_name
            self.owner_kind = resp.owner_kind
            if resp.carrier:
                self.carrier_name = resp.carrier.name
                self.kind = resp.carrier.type
            self.is_valid = self.owner_kind is not None
            self.sms_allowed = self.kind in ["mobile", "voip"]
        self.save()

    @staticmethod
    def lookup(number, country="US"):
        number = phone_util.normalize(number, country)
        info = PhonenumberInfo.objects.filter(number=number).last()
        if info is None:
            info = PhonenumberInfo(number=number)
            if phone_util.isValid(number):
                info.refresh()
        return info

    @staticmethod
    def normalize(number, country="US"):
        return phone_util.normalize(number, country)






import requests
import threading
from concurrent import futures

from redis import ConnectionPool, StrictRedis
from ws4redis import settings as ws4redis_settings

redis_connection_pool = ConnectionPool(**ws4redis_settings.WS4REDIS_CONNECTION)

def getRedisClient():
    return StrictRedis(connection_pool=redis_connection_pool)

from .models import redis, Task, TASK_STATE_STARTED
from .transports import email, http, sftp, sms, s3
from rest import helpers
from rest.log import getLogger
from rest import settings
from objict import nobjict
from datetime import datetime, timedelta
import time


TQ_WORKERS = settings.get("TQ_WORKERS", 4)
TQ_SUBSCRIBE = settings.get("TQ_SUBSCRIBE", [])

logger = None

# use threads or processes
# because tasks are general heavy IO bound threads are probably more efficient
USE_THREADS = True

import multiprocessing
CPU_COUNT = multiprocessing.cpu_count()

if not USE_THREADS:
    # only use the number of cpus we have if using process
    TQ_WORKERS = min(TQ_WORKERS, CPU_COUNT)

class WorkManager(object):
    def __init__(self, worker_count=TQ_WORKERS, subscribe_to=TQ_SUBSCRIBE, **kwargs):
        self.worker_count = worker_count
        self.subscribe_to = subscribe_to
        self.logger = kwargs.get("logger", None)
        self.service = kwargs.get("service", None)
        self.client = None
        self.pubsub = None
        self._scheduled_tasks = {}
        self._running_count = 0
        self._pending_count = 0
        self.is_running = False
        self.host_channel = f"tq:host:{settings.HOSTNAME}"
        self.lock = threading.RLock()
        if not self.logger:
            self.logger = getLogger("root", filename="tq_worker.log")
        self.logger.info("starting manager, workers: {}".format(self.worker_count))
        self.logger.info("handling: {}".format(self.subscribe_to))
        if USE_THREADS:
            self._pool = futures.ThreadPoolExecutor(max_workers=self.worker_count)
        else:
            self._pool = futures.ProcessPoolExecutor(max_workers=self.worker_count)

    def updateCounts(self):
        self.logger.info("running: {} --- pending: {}".format(self._running_count, self._pending_count))

    def addTask(self, task):
        if task.is_stale:
            self.logger.warning("task({}) is now stale".format(task.id))
            task.failed("stale")
            return
        if task.id in self._scheduled_tasks:
            self.logger.error("task({}) is alrleady scheduled".format(task.id))
            return
        task.manager = self
        with self.lock:
            task.worker_running = False
            self._scheduled_tasks[task.id] = task
            self._pending_count += 1
            task.future = self._pool.submit(self.on_run_task, task)
        self.updateCounts()

    def addEvent(self, event):
        # self.logger.info("processing event", event)
        if event.type == "subscribe":
            # confirmation we subscribed
            self.logger.info("succesfully subscribed to: {}".format(event.channel))
            return

        self.logger.info(f"new_event@{event.channel}")

        if event.channel == "tq_restart":
            self.restart()
            return
        
        if event.data:
            event.data = nobjict.fromJSON(event.data)
        
        if event.channel == self.host_channel:
            self.on_host_event(event)
            return

        try:
            task = Task.FromEvent(event)
        except Exception:
            # this most likely means the db connection is broken!
            self.logger.exception("FromEvent")
            # recommend a restart
            self.logger.warning("db error? forcing a restart...")
            self.restart()
            return

        if not task:
            self.logger.warning("event has no task", event)
            return
        if event.channel == "tq_cancel":
            self.logger.info("cancel request received")
            try:
                self.cancelTask(task, event.data.reason)
            except Exception:
                self.logger.exception("during cancelTask")
            return
        self.addTask(task)

    def cancelTask(self, task, reason=None):
        cached_task = self._scheduled_tasks.get(task.id, None)
        if not cached_task:
            # task is not scheduled
            self.logger.warning("canceling non scheduled task({})".format(task.id))
            task.state = -2
            task.reason = reason
            task.save()
            return
        if not hasattr(cached_task, "future"):
            self.removeTask(cached_task)
            self.logger.error("task has no future!")
            return
        task = cached_task
        if task.future.running():
            # right now we don't support canceling a running task but we will try!
            self.logger.warning("attempting to stop running task({})".format(task.id))
            if self.killWorker(task._thread_id):
                time.sleep(2.0)
                if task.future.done():
                    self.logger.info("succesfully killed task({}@{})".format(task.id, task._thread_id))
                    task.state = -2
                    task.reason = reason
                    task.save()
                else:
                    self.logger.warning("failed to kill worker")
            else:
                self.logger.warning("failed to kill worker")
        else:
            if task.future.cancel():
                self.logger.info("succesfully canceled task({})".format(task.id))
                task.state = -2
                task.reason = reason
                task.save()
                self.removeTask(task)
                return
            else:
                self.logger.error("failed to cancel task({})".format(task.id))

    def killWorker(self, thread_id):
        import ctypes
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, ctypes.py_object(SystemExit))
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
            return False
        return True

    def removeTask(self, task):
        try:
            with self.lock:
                del self._scheduled_tasks[task.id]
                self._pending_count -= 1
                self.updateCounts()
        except Exception:
            pass

    def processBacklog(self):
        tasks = Task.objects.filter(state__in=[0,1,2])
        for task in tasks:
            if task.channel in self.subscribe_to:
                if task.cancel_requested:
                    self.logger.info("task has cancel request {}".format(task.id))
                    task.state = -2
                    if not task.reason:
                        task.reason = "task canceled"
                    task.save()
                    continue
                self.logger.debug("resubmitting job {}".format(task.id))
                self.addTask(task)
            else:
                self.logger.warning("ignore job {}:{}".format(task.id, task.channel))

    def _on_webrequest(self, task):
        # self.logger.debug("starting web request to: {}".format(task.data.url))
        # we need to copy the data and remove the url
        if task.fname.upper() == "GET":
            resp = http.GET(task)
        else:
            resp = http.POST(task)
        if resp:
            task.completed()
        elif task.state == 1:
            # slowly backoff retries
            task.retry_later(from_now_secs=(task.attempts-1) * 120)

    def _on_hookrequest(self, task):
        # we need to copy the data and remove the url
        if task.model == "tq_email_request":
            resp = email.SEND(task)
        elif task.model == "tq_sms_request":
            resp = sms.SEND(task)
        elif task.model == "tq_sftp_request":
            resp = sftp.SEND(task)
        elif task.model == "tq_s3_request":
            resp = s3.UPLOAD(task)
        if resp:
            task.completed()
        elif task.state == 1:
            task.retry_later(from_now_secs=task.attempts * 60)
        else:
            task.failed()

    def on_host_event(self, event):
        action = event.data.action
        if action == "ping":
            redis.publish(
                event.data.response_channel,
                dict(action="pong", hostname=settings.HOSTNAME))
        elif action == "restart":
            redis.publish(
                event.data.response_channel,
                dict(action="restarting", hostname=settings.HOSTNAME))
            self.restart() 
        elif action == "get_stats":
            data = nobjict(action="stats", hostname=settings.HOSTNAME)
            data.uptime = time.time() - self.started_at
            data.subscribed = self.subscribe_to
            data.running = self._running_count
            data.pending = self._pending_count
            data.scheduled = len(self._scheduled_tasks.keys())
            redis.publish(event.data.response_channel, data)

    def on_task_started(self, task):
        with self.lock:
            task.worker_running = True
            self._running_count += 1
            self._pending_count -= 1
            self.updateCounts()

    def on_task_ended(self, task):
        with self.lock:
            task.worker_running = False
            del self._scheduled_tasks[task.id]
            self._running_count -= 1
            self.updateCounts()

    def on_run_task(self, task):
        self.logger.info("running task({})".format(task.id))

        # check if its canceled
        try:
            self.on_task_started(task)
            task.refresh_from_db()
            task._thread_id = threading.current_thread().ident
            self.logger.debug("running on thread:{}".format(task._thread_id))
            if task.state not in [0, 1, 2, 10] or task.cancel_requested:
                self.logger.info("task({}) was canceled?".format(task.id))
                self.on_task_ended(task)
                return

            if task.is_stale:
                self.logger.warning("task({}) is now stale".format(task.id))
                task.failed("stale")
                self.on_task_ended(task)
                return
        except Exception as err:
            self.logger.exception(err)
            return

        handler = None
        if task.model == "tq_web_request":
            handler = self._on_webrequest
        elif task.model in ["tq_sftp_request", "tq_email_request", "tq_sms_request", "tq_s3_request"]:
            handler = self._on_hookrequest
        else:
            try:
                handler = task.getHandler()
            except Exception as err:
                self.logger.exception(err)
                task.log_exception(err)
                task.failed(str(err))
                self.on_task_ended(task)
                return

        if handler is None:
            self.logger.error("failed to find handler: task({})".format(task.id))
            task.failed("failed to find handler")
            self.on_task_ended(task)
            return

        # self.logger.debug("task.started()")
        task.started()
        try:
            # self.logger.debug("task({}) calling handler".format(task.id))
            handler(task)
            if task.state == TASK_STATE_STARTED:
                task.completed()
            # self.logger.debug("task({}) handler finished".format(task.id))
        except Exception as err:
            self.logger.exception(err, "task({}) had exception".format(task.id))
            task.log_exception(err)
            if "connection already closed" in str(err).lower():
                # this is a nasty little bug in django when forking django db connections
                # we will schedule the task to retry later
                task.retry_later()
                # let us try and close db connections?
                hack_closeDjangoDB()
                # or should we restart the task queue?
            else:
                task.failed(str(err))
        except SystemExit:
            self.logger.error("task({}) was killed".format(task.id))
        finally:
            self.on_task_ended(task)
        self.logger.info("task({}) finished with state {}".format(task.id, task.state))

    def run_forever(self):
        self.logger.info("starting work manager...")
        self.__open()
        self.logger.info("listening for incoming events...")
        while self.is_running:
            for event in self.pubsub.listen():
                if self.is_running:
                    # self.logger.debug("new event", event)
                    try:
                        event = nobjict.fromdict(event)
                        event.channel = helpers.toString(event.channel)
                        self.addEvent(event)
                    except Exception as err:
                        self.logger.exception(err)
        self.__close()

    def restart(self):
        if self.service:
            self.is_running = False
            self.stop(timeout=30.0)
            self.service.restart()

    def __open(self):
        if self.client is not None:
            return
        self.started_at = time.time()
        self.is_running = True
        self.client = getRedisClient()
        self.pubsub = self.client.pubsub()
        self.__subscribe()

    def __subscribe(self):
        if self.host_channel not in self.subscribe_to:
            self.subscribe_to.append(self.host_channel)
        for key in self.subscribe_to:
            self.logger.info("subscribing to: {}".format(key))
            self.pubsub.subscribe(key)
        self.pubsub.subscribe("tq_cancel")
        self.pubsub.subscribe("tq_restart")
        self.client.sadd("tq:host:online", settings.HOSTNAME)

    def __unsubscribe(self):
        self.client.srem("tq:host:online", settings.HOSTNAME)
        for key in self.subscribe_to:
            self.pubsub.unsubscribe(key)
        self.pubsub.unsubscribe("tq_cancel")
        self.pubsub.unsubscribe("tq_restart")

    def __close(self):
        if self.client is None:
            return

        self.__unsubscribe()
        self.client = None
        self.pubsub = None
        self.logger.info("closed")

    def stop(self, timeout=30.0):
        self.updateCounts()
        self.logger.info("stopping, canceling pending tasks...")
        self.is_running = False
        # we need to cancel all futures not running:
        try:
            redis.publish("tq_cancel", {"pk":1})
            with self.lock:
                self.updateCounts()
                for key, task in list(self._scheduled_tasks.items()):
                    if not hasattr(task, "future"):
                        continue
                    if not task.future.running():
                        task.future.cancel()
                self.updateCounts()
        except Exception as err:
            self.logger.exception(err)

        self.logger.info("waiting for {} running tasks, timeout: {}".format(self._running_count, timeout))
        time.sleep(1.0)
        self.__close()
        timeout_at = time.time() + timeout
        while self._running_count > 0 and time.time() < timeout_at:
            # we are waiting for all jobs to finish
            time.sleep(1.0)

        self.updateCounts()
        if self._running_count:
            self.logger.error("timedout waiting for long running tasks, stopping anyway")
            # lets set all to failed that are running
            for pk, task in list(self._scheduled_tasks.items()):
                if task.worker_running:
                    if task.current_runtime >= 300:
                        task.failed("killed because task is taking to long to run")
                    else:
                        task.retry_later("worker engine restarting")
        else:
            self.logger.info("all tasks complete and workers stopped!")


def hack_closeDjangoDB():
    from django.db import connections
    for conn in connections.all():
        conn.close_if_unusable_or_obsolete()





"""
This is where you can put handlers for running async background tasks

Task.Publish("myapp", "on_tq_test")
"""
# from datetime import datetime, timedelta
# from auditlog.models import PersistentLog
# from django.conf import settings
from incident import models as ia
from account import models as account
from location.models import GeoIP
from rest import helpers
from rest import settings
from inbox import models as inbox
import time
from datetime import datetime, timedelta


POSIX_BODY_BREAK = "\n\n\n"
WIN_BODY_BREAK = "\r\n\r\n\r\n"


def new_event(task):
    data = task.data
    if "hostname" in data.metadata:
        data.hostname = data.metadata.hostname
    if "details" in data.metadata:
        data.details = data.metadata.details
    if "component" in data.metadata:
        data.component = data.metadata.component
    if "component_id" in data.metadata:
        data.component_id = data.metadata.component_id
    if "ip" in data.metadata:
        data.reporter_ip = data.metadata.ip
    ia.Event.createFromDict(None, task.data)
    task.completed()


def firewall_block(task):
    # Task.Publish("incident", "firewall_block", {ip:"x.x.x.x"}, channel="tq_broadcast")
    helpers.blockIP(task.data.ip)
    task.log(f"{settings.HOSTNAME} - {task.data.ip} BLOCKED")
    task.completed()


def firewall_unblock(task):
    # Task.Publish("incident", "firewall_unblock", {ip:"x.x.x.x"}, channel="tq_broadcast")
    helpers.unblockIP(task.data.ip)
    task.log(f"{settings.HOSTNAME} - {task.data.ip} UNBLOCKED")
    task.completed()


def on_incoming_email(task):
    # dict(pk=msg.pk, to_email=msg.to_email, from_email=msg.from_email)
    msg = inbox.Message.objects.filter(pk=task.data.pk).last()
    if msg is None:
        time.sleep(2.0)
        msg = inbox.Message.objects.filter(pk=task.data.pk).last()
        if msg is None:
            task.failed("could not find message")
            return
    # first we attempt to find the incident id from the subject
    if "#" not in msg.subject:
        task.failed("could not find incident id in subject")
        return
    pk = msg.subject[msg.subject.find("#")+1:]
    if ' ' in pk:
        pk = pk[:pk.find(' ')]
    obj = ia.Incident.objects.filter(pk=int(pk)).last()
    if obj is None:
        task.failed(f"could not find incident #{pk}")
        return
    member = account.Member.objects.filter(email=msg.from_email).last()
    username = msg.from_email
    if member is not None:
        username = member.username
    body = msg.body
    if POSIX_BODY_BREAK in body:
        body = body[:body.find(POSIX_BODY_BREAK)]
    elif WIN_BODY_BREAK in body:
        body = body[:body.find(WIN_BODY_BREAK)]
    lines = body.split("\n")
    note = []
    for line in lines:
        if line.startswith(">"):
            break
        action = line.strip().lower()
        if action in ["close", "closed", "resolve", "resolved", "handled"]:
            obj.state = ia.INCIDENT_STATE_RESOLVED
            obj.save()
            msg = f"resolved by {username}"
            note.append(msg)
            obj.logHistory("history", msg, notify=False)
        elif action in ["open", "accept"]:
            obj.state = ia.INCIDENT_STATE_OPENED
            obj.save()
            msg = f"opened by {username}"
            note.append(msg)
            obj.logHistory("history", msg, notify=False)
        elif action in ["ignore"]:
            obj.state = ia.INCIDENT_STATE_IGNORE
            obj.save()
            msg = f"ignored by {username}"
            note.append(msg)
            obj.logHistory("history", msg, notify=False)
        elif action in ["pause"]:
            obj.state = ia.INCIDENT_STATE_PAUSE
            obj.save()
            msg = f"paused by {username}"
            note.append(msg)
            obj.logHistory("history", msg, notify=False)
        else:
            note.append(line)
    obj.logHistory("email", "\n".join(note), member=member)


def run_cleanup(task):
    stale = datetime.now() - timedelta(days=90)
    # delete all ossec alerts older then 90 days
    count = ia.ServerOssecAlert.objects.filter(created__lte=stale).delete()
    if count:
        task.log(f"deleted {count} old ServerOssecAlert")
    # delete all events older then 90 days
    count = ia.Event.objects.filter(created__lte=stale).delete()
    if count:
        task.log(f"deleted {count} old Events")

    count = GeoIP.removeDuplicates()
    if count:
        task.log(f"deleted {count} duplicate IPs")




from django.template import Context
from django.template import Template
from django.db import models
from rest.models import RestModel

try:
    import css_inline
    inliner = css_inline.CSSInliner(remove_style_tags=True)
except Exception:
    inliner = None


class MailTemplate(models.Model, RestModel):
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)
    group = models.ForeignKey("account.Group", blank=True, null=True, default=None, related_name="templates", on_delete=models.CASCADE)
    name = models.CharField(max_length=255, db_index=True)
    kind = models.CharField(max_length=124, default="email")
    template = models.TextField()

    def render(self, context):
        template = Template(self.template)
        if context is None:
            context = {}
        context = Context(context)
        if inliner is not None:
            return inliner.inline(template.render(context))
        return template.render(context)


import os
from .settings_helper import settings
from rest.crypto.privpub import PrivatePublicEncryption


DECRYPTER_KEY_FILE = os.path.join(settings.PROJECT_ROOT, "config", "decrypt_key.pem")
DECRYPTER = None
if not os.path.exists(DECRYPTER_KEY_FILE):
    print(("WARNING, failed to load decrypter!!! {}".format(DECRYPTER_KEY_FILE)))
    DECRYPTER_KEY_FILE = os.path.join(os.path.dirname(os.path.dirname(settings.ROOT)), "config", "decrypt_key.pem")

if os.path.exists(DECRYPTER_KEY_FILE):
    DECRYPTER = PrivatePublicEncryption(private_key_file=DECRYPTER_KEY_FILE)


ENCRYPTER_KEY_FILE = os.path.join(settings.PROJECT_ROOT, "config", "encrypt_key.pem")
ENCRYPTER = None
if os.path.exists(ENCRYPTER_KEY_FILE):
    ENCRYPTER = PrivatePublicEncryption(private_key_file=ENCRYPTER_KEY_FILE)



import http.client
from . import httplib2
from urllib.parse import urlencode
from urllib.request import Request, urlopen
import json


from medialib.youtube.apiclient import discovery
from medialib.youtube.oauth2client.client import AccessTokenCredentials

import logging

from medialib.render import render_utils

# in module
from . import upload


httplib2.RETRIES = 1

VALID_PRIVACY_STATUSES = ("public", "private", "unlisted")


def getService(access_token):
    """
    Retrieves a usable YouTubeService via just an access_token
    """
    if hasattr(access_token, "logger"):
        return access_token

    credentials = AccessTokenCredentials(access_token, 'medialib-311/1.0')
    http = httplib2.Http()
    http = credentials.authorize(http)
    # Construct the service object for the interacting with the YouTube Data API.
    service = discovery.build('youtube', 'v3', http=http)
    service.logger = logging.getLogger("youtube")
    return service

def testToken(access_token):
    data = getUserDetails(access_token)
    return data and "email" in data

def refreshToken(refresh_token, client_id, client_secret):
    data = urlencode({
        'refresh_token': refresh_token,
        'client_id': client_id,
        'client_secret': client_secret,
        'grant_type': 'refresh_token'
        })

    request = Request('https://accounts.google.com/o/oauth2/token', data)
    try:
        data = json.loads(urlopen(request).read())
        return data
    except (ValueError, IOError):
        pass
    return None

def hasChannel(access_token):
    try:
        channels = listChannels(access_token)
    except:
        return False
    for ch in channels:
        if "snippet" in ch and "title" in ch["snippet"] and len(ch["snippet"]["title"]):
            return True
    return False

def getUserDetails(access_token):
    data = urlencode({'oauth_token': access_token, 'alt': 'json'})
    request = Request('https://www.googleapis.com/userinfo/v2/me?' + data, headers={'Authorization': data})
    try:
        data = json.loads(urlopen(request).read())
        return data
    except (ValueError, IOError):
        pass
    return None

def getVideos(access_token, video_ids=""):
    service = getService(access_token)
    # Retrieve the contentDetails part of the channel resource for the
    # authenticated user's channel.
    videos_response = service.videos().list(
        id=video_ids,
        part="id, status, statistics"
    ).execute()

    if not videos_response["items"]:
        service.logger.warning("no video results returned")
        return None

    return videos_response

def listChannels(access_token):
    service = getService(access_token)
    channels_response = service.channels().list(
      mine=True,
      part="id, snippet"
    ).execute()
    return channels_response["items"]

def listVideos(access_token, video_ids=""):
    service = getService(access_token)
    # Retrieve the contentDetails part of the channel resource for the
    # authenticated user's channel.
    channels_response = service.channels().list(
        mine=True,
        part="contentDetails"
    ).execute()

def uploadVideoFile(access_token, videofile, title, description, category="22", keywords="", privacy="public"):
    """
    Upload a video file with just an access_token
    """
    service = getService(access_token)
    return upload.resumable_upload(service, videofile, title, description, category, keywords, privacy)


def uploadMediaItem(access_token, media_item, renditiondef, params=None):
    """
    Upload a media item
    """
    from medialib.models import MediaItemRendition
    # get the original video file to transfer
    # TODO get straight from S3 bucket???
    original = render_utils.get_rendition(media_item, 'original')
    title = media_item.name
    description = media_item.description
    if params and "description" in params:
        description = params["description"]

    # upload the actual file
    response = uploadVideoFile(access_token, original.filename, title, description)

    # now update the rendition
    rendition = MediaItemRendition(mediaitem=media_item, name=renditiondef.name, use=renditiondef.use, rendition_definition=renditiondef, bytes=0, url='{0}'.format(response["id"]))
    rendition.save()

    return rendition



from django.conf import settings
from rest import UberDict
import phonenumbers

try:
    import twilio
    import twilio.rest
except ImportError:
    pass

client = twilio.rest.Client(settings.TWILIO_SID, settings.TWILIO_AUTH_TOKEN)


def lookup(number, country="US", name_lookup=True):
    extra_info = ['carrier']
    if name_lookup:
        extra_info.append("caller-name")
    number = normalize(number, country)
    output = UberDict(success=False)
    try:
        resp = client.lookups.v1.phone_numbers(number).fetch(type=extra_info)
    except Exception as err:
        output.reason = str(err)
        return output
    output.success = True
    output.number = number
    output.carrier = UberDict.fromdict(resp.carrier)
    if resp.caller_name:
        output.owner_name = resp.caller_name.get("caller_name", None)
        output.owner_kind = resp.caller_name.get("caller_type", None)
    return output


def sendSMS(to_num, from_num, msg, country="US"):
    to_num = normalize(to_num, country)
    from_num = normalize(from_num, country)
    tmsg = client.messages.create(body=msg, to=to_num, from_=from_num)
    return tmsg


def find(text, country="US"):
    # finds any phone numbers in the text blob
    numbers = []
    for match in phonenumbers.PhoneNumberMatcher(text, country):
        numbers.append(normalize(match.number, country))
    return numbers


def isValid(number, country="US"):
    return phonenumbers.is_valid_number(phonenumbers.parse(number, country))


def normalize(raw_phone, country="US"):
    # turns a number like 202-413-2409 into +12024132409
    return convert_to_e164(raw_phone, country)


def convert_to_e164(raw_phone, country="US"):
    if not raw_phone:
        return

    if raw_phone[0] == '+':
        # Phone number may already be in E.164 format.
        parse_type = None
    else:
        # If no country code information present, assume it's a US number
        parse_type = country

    phone_representation = phonenumbers.parse(raw_phone, parse_type)
    return phonenumbers.format_number(
        phone_representation,
        phonenumbers.PhoneNumberFormat.E164)



from django.db import models as dm
from rest import fields as rf
from rest import helpers as rh
from objict import objict


class ModelCache(dm.Model):
    class Meta:
        abstract = True

    modified = dm.DateTimeField(auto_now=True, editable=True, db_index=True)

    component = dm.SlugField(max_length=200, null=True, blank=True, default=None, db_index=True)
    component_id = dm.IntegerField(null=True, blank=True, default=None, db_index=True)

    query = dm.CharField(max_length=255, null=True, blank=True, default=None, db_index=True)

    cache = rf.JSONField()




# Copyright (C) 2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A keyring based Storage.

A Storage for Credentials that uses the keyring module.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import keyring
import threading

from .client import Storage as BaseStorage
from .client import Credentials


class Storage(BaseStorage):
  """Store and retrieve a single credential to and from the keyring.

  To use this module you must have the keyring module installed. See
  <http://pypi.python.org/pypi/keyring/>. This is an optional module and is not
  installed with oauth2client by default because it does not work on all the
  platforms that oauth2client supports, such as Google App Engine.

  The keyring module <http://pypi.python.org/pypi/keyring/> is a cross-platform
  library for access the keyring capabilities of the local system. The user will
  be prompted for their keyring password when this module is used, and the
  manner in which the user is prompted will vary per platform.

  Usage:
    from oauth2client.keyring_storage import Storage

    s = Storage('name_of_application', 'user1')
    credentials = s.get()

  """

  def __init__(self, service_name, user_name):
    """Constructor.

    Args:
      service_name: string, The name of the service under which the credentials
        are stored.
      user_name: string, The name of the user to store credentials for.
    """
    self._service_name = service_name
    self._user_name = user_name
    self._lock = threading.Lock()

  def acquire_lock(self):
    """Acquires any lock necessary to access this Storage.

    This lock is not reentrant."""
    self._lock.acquire()

  def release_lock(self):
    """Release the Storage lock.

    Trying to release a lock that isn't held will result in a
    RuntimeError.
    """
    self._lock.release()

  def locked_get(self):
    """Retrieve Credential from file.

    Returns:
      oauth2client.client.Credentials
    """
    credentials = None
    content = keyring.get_password(self._service_name, self._user_name)

    if content is not None:
      try:
        credentials = Credentials.new_from_json(content)
        credentials.set_store(self)
      except ValueError:
        pass

    return credentials

  def locked_put(self, credentials):
    """Write Credentials to file.

    Args:
      credentials: Credentials, the credentials to store.
    """
    keyring.set_password(self._service_name, self._user_name,
                         credentials.to_json())

  def locked_delete(self):
    """Delete Credentials file.

    Args:
      credentials: Credentials, the credentials to store.
    """
    keyring.set_password(self._service_name, self._user_name, '')



import sys

from functools import wraps
from django.urls import URLResolver
from django.urls import re_path, path
from django.shortcuts import Http404
from django.http import HttpResponseRedirect
from django.utils.cache import patch_cache_control, add_never_cache_headers, patch_vary_headers

from rest import settings
from rest.views import restStatus, restPermissionDenied
from rest.errors import RestError, PermissionDeniedException
from rest import helpers as rh
import metrics
import incident

from datetime import datetime
from auditlog.models import PersistentLog

import importlib
import pkgutil
import threading
import traceback

REST_METRICS = settings.get("REST_METRICS", False)
REST_METRICS_BY_ENDPOINT = settings.get("REST_METRICS_BY_ENDPOINT", False)
REST_METRICS_IGNORE = settings.get("REST_METRICS_IGNORE", [])
REST_MODULE_NAME = settings.get("REST_MODULE_NAME", "rpc")


# background task (no return)
def rest_async(func):
    """
    Execute the function asynchronously in a separate thread
    """
    @wraps(func)
    def inner(*args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
    return inner


def postpone(function):
    def decorator(*args, **kwargs):
        t = threading.Thread(target=function, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
    return decorator


#
# Annotate a view with the URL that points to it
#
def rest_error_catcher(func, request, *args, **kwargs):
    try:
        if REST_METRICS:
            metrics.metric("rest_calls")
            if REST_METRICS_BY_ENDPOINT:
                slug_path = request.path.replace("/", "__")
                metrics.metric(f"rest_call_{slug_path}", category="rest_calls")
        return func(request, *args, **kwargs)
    except PermissionDeniedException as err:
        rh.log_error("permission denied: '{}' for {} {}:{}".format(err, request.user, request.method, request.path))
        return restStatus(request, False, error=err.reason, error_code=err.code)
    except RestError as err:
        rh.log_exception("REST ERROR", request.path, err.reason)
        if settings.get("REST_ERROR_METRICS", True):
            metrics.metric("rest_errors")
        if settings.get("REST_ERROR_EVENTS", True):
            incident.exception_event(
                request, f"{err.code}: {err.reason}", level=7,
                category="rest_error",
                reason=err.reason, code=err.code)
        return restStatus(request, False, error=err.reason, error_code=err.code)
    except Exception as err:
        rh.log_exception("REST EXCEPTION", request.path)
        if settings.get("REST_ERROR_METRICS", True):
            metrics.metric("rest_errors")
        try:
            body = request.body.decode('utf-8')
        except Exception:
            body = request.DATA.asDict()
        PersistentLog.logException(body, request=request, component="rest", action="error")
        stack = str(traceback.format_exc())
        incident.exception_event(request, err, body, stack)
        return restStatus(request, False, error=str(err), stack=stack)
    return restStatus(request, False)


def dispatcher(request, *args, **kwargs):
    module = kwargs.pop('__MODULE')
    pattern = kwargs.pop('__PATTERN')
    method = request.method
    if request.method == 'HEAD':
        method = 'GET'
    key = pattern + '__' + method
    if key in module.urlpattern_methods:
        return rest_error_catcher(module.urlpattern_methods[key], request, *args, **kwargs)
    method = "ALL"
    key = pattern + '__' + method
    if key in module.urlpattern_methods:
        return rest_error_catcher(module.urlpattern_methods[key], request, *args, **kwargs)
    # print module.urlpattern_methods
    return restStatus(request, False, error="endpoint not found", error_code=404)


def _load_module(mod):
    if pkgutil.find_loader(mod) is not None:
        return importlib.import_module(mod)
    return None


def _url_method(pattern, method=None, *args, **kwargs):
    """
    Register a view handler for a specific HTTP method
    """
    caller_filename = sys._getframe(2).f_code.co_filename
    module = None
    for m in list(sys.modules.values()):
        if m and '__file__' in m.__dict__ and m.__file__ is not None and m.__file__.startswith(caller_filename):
            module = m
            break

    def _wrapper(f):
        new_pattern = True
        if module:
            rpc_root_module = module
            if module.__name__.count('.') > 1:
                # this means we are not in root
                # print module.__name__
                root_name = module.__name__.split('.')[0]
                # print "importing {0}.rpc".format(root_name)
                rmodule = _load_module(f"{root_name}.{REST_MODULE_NAME}")
                if rmodule is not None:
                    rpc_root_module = rmodule
            # print "{0}/{1}".format(rpc_root_module.__name__, pattern)
            elif not module.__name__.endswith(f".{REST_MODULE_NAME}") and module.__name__.count('.'):
                # print module.__name__
                root_name = module.__name__.split('.')[0]
                # print "importing {0}.rpc".format(root_name)
                rmodule = _load_module(f"{root_name}.{REST_MODULE_NAME}")
                if rmodule is not None:
                    rpc_root_module = rmodule
            lmethod = method
            if lmethod is None:
                lmethod = "ALL"
            if 'urlpatterns' not in rpc_root_module.__dict__:
                rpc_root_module.urlpatterns = []
            if lmethod and 'urlpattern_methods' not in rpc_root_module.__dict__:
                rpc_root_module.urlpattern_methods = {}
            elif lmethod and pattern + '__' in rpc_root_module.urlpattern_methods:
                new_pattern = False

            if lmethod:
                rpc_root_module.urlpattern_methods[pattern + '__' + lmethod] = f

            if new_pattern:
                if lmethod:
                    func = dispatcher
                    func.csrf_exempt = True
                    rpc_root_module.urlpattern_methods[pattern + '__'] = True
                    if 'kwargs' not in kwargs:
                        kwargs['kwargs'] = {}
                    kwargs['kwargs']['__MODULE'] = rpc_root_module
                    kwargs['kwargs']['__PATTERN'] = pattern
                else:
                    func = f
                if not isinstance(pattern, str):
                    rh.log_print("NOT A STRING", pattern)
                if pattern.startswith("^") or pattern.endswith("$"):
                    rpc_root_module.urlpatterns += [re_path(pattern, func, *args, **kwargs)]
                else:
                    rpc_root_module.urlpatterns += [path(pattern, func, *args, **kwargs)]
                    if not pattern.endswith("/") and not pattern.endswith(">"):
                        rpc_root_module.urlpatterns += [path(pattern + "/", func, *args, **kwargs)]
            f.__url__ = (lmethod, pattern)
            f.csrf_exempt = True
        return f
    _wrapper.caller_filename = "{0}/{1}".format(module.__name__, pattern)
    return _wrapper


def url(pattern, *args, **kwargs):
    """
    Usage:
    @url(r'^users$')
    def get_user_list(request):
        ...
    """
    return _url_method(pattern, *args, **kwargs)


def urlGET(pattern, *args, **kwargs):
    """
    Register GET handler for url pattern
    """
    return _url_method(pattern, 'GET', *args, **kwargs)


def urlPUT(pattern, *args, **kwargs):
    """
    Register PUT handler for url pattern
    """
    return _url_method(pattern, 'PUT', *args, **kwargs)


def urlPOST(pattern, *args, **kwargs):
    """
    Register POST handler for url pattern
    """
    return _url_method(pattern, 'POST', *args, **kwargs)


def urlPOST_NOCSRF(pattern, *args, **kwargs):
    """
    Register POST handler for url pattern
    """
    return _url_method(pattern, 'POST', *args, **kwargs)


def urlDELETE(pattern, *args, **kwargs):
    """
    Register DELETE handler for url pattern
    """
    return _url_method(pattern, 'DELETE', *args, **kwargs)


#
# Continue the @url decorator pattern into sub-modules, if desired
#

def include_urlpatterns(regex, module):
    """
    Usage:

    # in top-level module code:
    urlpatterns = include_urlpatterns(r'^profile/', 'apps.myapp.views.profile')
    """
    return [URLResolver(regex, module)]


#
# patched django decorators check if return is httpresponse
#
def cache_control(**kwargs):
    def _cache_controller(view_func):
        @wraps(view_func)
        def _cache_controlled(request, *args, **kw):
            response = view_func(request, *args, **kw)
            if hasattr(response, 'has_header'):  # check if response is httpresponse
                patch_cache_control(response, **kwargs)
            return response
        return _cache_controlled
    return _cache_controller


def never_cache(view_func):
    @wraps(view_func)
    def _wrapped_view_func(request, *args, **kwargs):
        response = view_func(request, *args, **kwargs)
        if hasattr(response, 'has_header'):  # check if response is httpresponse
            add_never_cache_headers(response)
        return response
    return _wrapped_view_func


def vary_on_headers(*headers):
    def decorator(func):
        @wraps(func)
        def inner_func(*args, **kwargs):
            response = func(*args, **kwargs)
            if hasattr(response, 'has_header'):  # check if response is httpresponse
                patch_vary_headers(response, headers)
            return response
        return inner_func
    return decorator


def vary_on_cookie(func):
    @wraps(func)
    def inner_func(*args, **kwargs):
        response = func(*args, **kwargs)
        if hasattr(response, 'has_header'):  # check if response is httpresponse
            patch_vary_headers(response, ('Cookie',))
        return response
    return inner_func


def force_ssl(func):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        if (not settings.DEBUG) and request and not request.is_secure():
            url = request.build_absolute_uri()
            return HttpResponseRedirect(url.replace('http://', 'https://'))

        response = func(request, *args, **kwargs)
        return response
    return inner_func


def login_required(func):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        if not request.user.is_authenticated:
            return restStatus(request, False, error="permission denied", error_code=401)
        return func(request, *args, **kwargs)
    return inner_func


def login_optional(func):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        return func(request, *args, **kwargs)
    return inner_func


def staff_required(func):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        if not request.user.is_authenticated or not request.user.is_staff:
            return restStatus(request, False, error="staff request denied", error_code=402)
        return func(request, *args, **kwargs)
    return inner_func


def superuser_required(func):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        if not request.user.is_authenticated or not request.user.is_superuser:
            return restStatus(request, False, error="admin request denied", error_code=402)
        return func(request, *args, **kwargs)
    return inner_func


class requires_params(object):
    def __init__(self, params):
        self.params = params

    def __call__(self, func):
        def inner_func(request=None, *args, **kwargs):
            if not request.DATA.hasRequired(self.params):
                return restPermissionDenied(request, error="missing required fields")
            return func(request, *args, **kwargs)
        return inner_func


class perm_required(object):
    def __init__(self, perms):
        self.perms = perms

    def __call__(self, func):
        def inner_func(request=None, *args, **kwargs):
            status, error, code = rh.requestHasPerms(request, self.perms)
            if not status:
                return restStatus(request, False, error=error, error_code=code)
            return func(request, *args, **kwargs)
        return inner_func


class post_perm_required(object):
    def __init__(self, perms):
        self.perms = perms

    def __call__(self, func):
        def inner_func(request=None, *args, **kwargs):
            if request.method == "post":
                status, error, code = rh.requestHasPerms(request, self.perms)
                if not status:
                    return restStatus(request, False, error=error, error_code=code)
            elif not request.user.is_authenticated:
                return restStatus(request, False, error="permission denied", error_code=401)
            return func(request, *args, **kwargs)
        return inner_func


def ip_whitelist(func, *args, **kwargs):
    @wraps(func)
    def inner_func(request=None, *args, **kwargs):
        request_ip = request.META['REMOTE_ADDR']
        if request_ip not in settings.AUTHORIZED_IPS:
            return restStatus(request, False, error="permission denied")
        return func(request, *args, **kwargs)
    return inner_func


def periodicCheckListHas(list_obj, has_value):
    if type(list_obj) is list:
        return has_value in list_obj
    return list_obj == has_value


PERIODIC_FUNCS = []
PERIODIC_EVERY_5_MINUTES = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55]
PERIODIC_EVERY_10_MINUTES = [0, 10, 20, 30, 40, 50]
PERIODIC_EVERY_15_MINUTES = [0, 15, 30, 45]
PERIODIC_EVERY_15_MINUTES_2 = [5, 20, 35, 50]
PERIODIC_EVERY_20_MINUTES = [5, 25, 45]
PERIODIC_EVERY_30_MINUTES = [15, 45]


def periodic(minute=None, hour=None, day=None, month=None, weekday=None, tz=None):
    """
    supports minute=5 or minute=[5,10,20]
    """
    def decorator(func):
        PERIODIC_FUNCS.append({
            "func": func,
            "minute": minute,
            "hour": hour,
            "day": day,
            "month": month,
            "weekday": weekday,
            "tz": tz
        })

        @wraps(func)
        def inner_func(force=False, verbose=False, now=None):
            if now is None:
                now = datetime.now()
            if force:
                return func(force, verbose, now)
            if tz:
                now = rh.convertToLocalTime(tz, now)
            # lets create our when
            if minute is not None and not periodicCheckListHas(minute, now.minute):
                return -22
            if hour is not None and not periodicCheckListHas(hour, now.hour):
                return -22
            if day is not None and not periodicCheckListHas(day, now.day):
                return -22
            if month is not None and not periodicCheckListHas(month, now.month):
                return -22
            if weekday is not None and not periodicCheckListHas(weekday, now.weekday()):
                return -22
            return func(force, verbose, now)
        inner_func.is_periodic = True
        return inner_func
    return decorator




# Generated by Django 4.1.4 on 2023-11-20 03:51

from django.db import migrations, models
import django.db.models.deletion
import rest.models.base
import rest.models.metadata


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0016_authsession_buid'),
        ('incident', '0010_incident_category_incident_component_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='Ticket',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('category', models.CharField(db_index=True, max_length=200)),
                ('priority', models.IntegerField(db_index=True, default=10)),
                ('title', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('status', models.CharField(db_index=True, default=None, max_length=32, null=True)),
                ('state', models.IntegerField(default=0)),
                ('component', models.CharField(db_index=True, default=None, max_length=200, null=True)),
                ('component_id', models.IntegerField(blank=True, db_index=True, default=None, null=True)),
                ('assigned_to', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tickets', to='account.member')),
                ('created_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member')),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tickets', to='account.group')),
                ('incident', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tickets', to='incident.incident')),
            ],
            bases=(models.Model, rest.models.base.RestModel, rest.models.metadata.MetaDataModel),
        ),
    ]



from ..render_utils import *
from ..exceptions import *

import os
from PIL import Image

PresetConfig = {
    'name':		'Validate Image',
    'description':	'Basic image validation',
    'stage':	1,
    'default_use':	'validate',
    'applicable_kinds': 'I',
    'parameters': {
    },
}

def image_is_animated_gif(image):
    # verify image format
    if image.format.lower() != 'gif':
        return False

    # verify GIF is animated by attempting to seek beyond the initial frame
    try:
        image.seek(1)
    except EOFError:
        return False
    else:
        return True

def render(item, renditiondef, params):
    original = get_rendition(item, 'original')

    try:
        im = Image.open(original.filename)
        if image_is_animated_gif(im):
            original.kind = "N"
        original.width = im.size[0]
        original.height = im.size[1]
        original.save()
    except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
        raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)



import ssl
import socket
from datetime import datetime
import concurrent.futures


def get_ssl_expiry_date(domain):
    """Get the SSL certificate expiry date of a domain."""
    # Connect to the domain
    context = ssl.create_default_context()
    conn = context.wrap_socket(socket.socket(socket.AF_INET), server_hostname=domain)
    conn.settimeout(3.0)
    
    try:
        conn.connect((domain, 443))
        ssl_info = conn.getpeercert()
        # Extract expiry date and convert to datetime object
        expiry_date = datetime.strptime(ssl_info['notAfter'], r'%b %d %H:%M:%S %Y %Z')
        return domain, expiry_date
    except Exception as e:
        return domain, f"Error fetching SSL info: {e}"
    finally:
        conn.close()


def check(*domains):
    """Check SSL expiry dates for a list of domains."""
    output = dict()
    if len(domains) == 1:
        if isinstance(domains[0], list):
            domains = domains[0]
        else:
            domain, result = get_ssl_expiry_date(domains[0])
            output[domain] = result
            return output

    with concurrent.futures.ThreadPoolExecutor(max_workers=len(domains)) as executor:
        futures = [executor.submit(get_ssl_expiry_date, domain) for domain in domains]
        for future in concurrent.futures.as_completed(futures):
            domain, result = future.result()
            output[domain] = result
    return output





from django.forms import ValidationError
import mimetypes
from io import StringIO
import base64
import binascii
import hashlib

EXT_MAP = {
    "V":['mpg', 'mpeg', 'mpe', 'mp4', 'm4v', 'mov', 'qt', '3gp', 'ogv', 'webm', 'flv', 'asf', 'asx', 'wmv', 'mpv', 'mkv', 'avi', 'dat', 'swf'],
    "I":['jpg', 'jpg', 'jpeg', 'gif', 'png', 'bmp', 'tif', 'tiff'],
    "T":['txt', 'html'],
}

KIND_MAP = {
    "video": "V",
    "image": "I",
    "http": "E"
}

def getFileExt(file):
    if hasattr(file, 'name'):
        return file.name.split(".")[-1].lower()
    return file.split(".")[-1].lower()

def guessBase64Kind(data):
    # look at first 16 bytes for ext
    sample = fromBase64ToBytes(data[:16]).lower() # data[:16]
    # sample = self.base64_data[:16].decode('base64').lower()
    for ext in ["png", "jpg", "jpeg", "jfif", "bmp", "gif", "tif"]:
        if toBytes(ext) in sample:
            if ext == "jfif":
                ext = "jpg"
            return ext
    return None

def getFileNameFromURL(url):
    return url.split("#")[0].split("?")[0].split("/")[-1]

def guessMediaKindByFile(file):
    if not hasattr(file, "name"):
        return guessMediaKindByName(file)
    if hasattr(file, 'content_type'):
        kind = guessMediaKindByContentType(file.content_type)
        if kind:
            return kind
    return guessMediaKindByFile(file.name)

def guessMediaKindByName(filename):
    kind = guessMediaKindByContentType(mimetypes.guess_type(filename))
    if kind:
        return kind
    # now lets look at our internal map
    ext = getFileExt(filename)
    for key in EXT_MAP:
        if ext in EXT_MAP[key]:
            return key
    return None

def guessMediaKindByContentType(content_type):
    # print content_type
    if type(content_type) is tuple:
        content_type = content_type[0]
    if not content_type:
        return None
    kind = content_type.split('/')[0]
    if kind == "video":
        return "V"
    elif kind == "image":
        return "I"
    elif kind == "text":
        return "T"
    return None

def guessMediaKind(file):
    kind = guessMediaKindByFile(file)
    if not kind:
        return '*'
    return kind

def validate_upload(file):
    kind = None
    kind = guessMediaKindByFile(file)
    if not kind:
        ext = getFileExt(file)
        return "*"
        # raise ValidationError("Invalid file type: .{0}".format(ext))
    return kind

# UTILS

def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value

def toBytes(value):
    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, bytearray):
        value = bytes(value)
    elif isinstance(value, (int, float)):
        value = str(value).encode('utf-8')
    return value

def toByteArray(value):
    if isinstance(value, bytearray):
        return value
    elif isinstance(value, str):
        return bytearray(value, 'utf-8')
    elif isinstance(value, bytes):
        return bytearray(value)
    return value

def toHex(value):
    return toBytes(value).hex().upper()

def hexToByteArray(value):
    return bytearray.fromhex(value)

def hexToString(value):
    return bytes.fromhex(value).decode('utf-8')

def toBase64(value):
    return base64.b64encode(value)

def fromBase64(value):
    return fromBase64ToString(value)

def fromBase64ToString(value):
    return toString(base64.b64decode(value))

def fromBase64ToBytes(value):
    return base64.b64decode(value)

def toMD5(*args):
    output = []
    for arg in args:
        output.append(hashlib.md5(toBytes(arg)).hexdigest())
    return "".join(output)



import sys
import re
import time
from datetime import date, datetime, timedelta
from .settings_helper import settings
import importlib
from django.db.models import Count, Q, Avg, Sum, Max, Min
from django.db.models.query import QuerySet
from io import StringIO
import base64
import binascii

from objict import objict
from .datem import *
from .log import getLogger
import getpass

from urllib.parse import urlparse


try:
    from fuzzywuzzy import fuzz
except:
    fuzz = None

# THIS NEEDS A NEW HOME
import shutil
import tempfile
import os
import random, string
import subprocess

ERROR_LOGGER = getLogger("errors", filename="error.log")
AUDIT_LOGGER = getLogger("auditlog", filename="auditlog.log")
DEBUG_LOGGER = getLogger("debug", filename="debug.log")
AUDITLOG_LOGGER_BY_USER = settings.get("AUDITLOG_LOGGER_BY_USER", False)
AUDITLOG_LOGGERS = settings.get("AUDITLOG_LOGGERS", {})
HELPER_CACHE = objict()

def getLoggerByRequest(request):
    logger = AUDIT_LOGGER

    if AUDITLOG_LOGGER_BY_USER and request and request.user.is_authenticated:
        return getLogger(request.user.username, "{}.log".format(request.user.username))
    elif AUDITLOG_LOGGERS:
        for path, key in list(AUDITLOG_LOGGERS.items()):
            # print(("{}: {}".format(path, key)))
            if request.path.startswith(path):
                return getLogger(key.split('.')[0], filename=key)
    return logger


def getActiveRequest():
    if not HELPER_CACHE.get_request:
        mw = importlib.import_module("rest.middleware")
        HELPER_CACHE.get_request = mw.get_request
    return HELPER_CACHE.get_request()


def requestHasPerms(request, perms, group=None):
    # third party auth models
    if request.auth_model is not None:
        if request.auth_model.hasPerm(perms):
            return True, None, None

    if not request.user.is_authenticated:
        return False, "auth required", 401

    if request.member.hasPerm(perms):
        return True, None, None
    if group is None and hasattr(request, "group"):
        group = request.group
    ms_perms = None
    if group:
        ms = request.member.getMembershipFor(group)
        if ms and ms.hasPerm(perms):
            return True, None, None
        if ms:
            ms_perms = ms.getPermissions()

    # rest_helpers.log_error(F"PERMISSION DENIED for {request.member.username}, {perms}", str(group), ms_perms)
    return False, "permission denied", 402

def importModule(name):
    return importlib.import_module(name)


def log_print(*args):
    getLoggerByRequest(getActiveRequest()).info(*args)


def log(*args):
    getLoggerByRequest(getActiveRequest()).info(*args)


def debug(*args):
    DEBUG_LOGGER.info(*args)


def getLogRequestHeader(request, no_data=True):
    extra = []
    if bool(request):
        if hasattr(request, "ip"):
            extra.append(request.ip)
        extra.append(request.build_absolute_uri())
        if hasattr(request, "member") and request.member:
            extra.append("user: {}".format(request.member.username))
        if not no_data and hasattr(request, "DATA"):
            extra.append(request.DATA.asDict())
    return extra


def log_exception(*args):
    request = getActiveRequest()
    extra = getLogRequestHeader(request)
    ERROR_LOGGER.exception(*extra, *args)
    getLoggerByRequest(request).exception(*extra, *args)


def log_error(*args):
    request = getActiveRequest()
    extra = getLogRequestHeader(request)
    ERROR_LOGGER.error(*extra, *args)
    getLoggerByRequest(request).error(*extra, *args)


def log_stack(*args):
    request = getActiveRequest()
    import inspect
    extra = []
    for frame in inspect.stack():
        file, line, func = frame[1:4]
        extra.append(f"{file}, line {line} in function {func}")
    ERROR_LOGGER.error(*extra, *args)
    getLoggerByRequest(request).error(*extra, *args)


def setLogModel(self, component, pk):
    # this method allows for updating the component to associate a log with
    if not pk or (self._log_pk == pk):
        return
    self._log_component = component
    self._log_pk = pk
    if hasattr(self, "request_log"):
        self.request_log.component = component
        self.request_log.pkey = pk
        self.request_log.save()
        

def getHostname():
    try:
        return os.uname()[1]
    except:
        pass
    return "unknown"

def getUsername():
    return getpass.getuser()

class TemporaryDirectory(object):
    """Context manager for tempfile.mkdtemp() so it's usable with "with" statement."""
    def __enter__(self):
        self.name = tempfile.mkdtemp()
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name)

class TemporaryFile(object):
    def __init__(self, suffix=".tmp"):
        temp_file = tempfile.NamedTemporaryFile(delete=False,   suffix=suffix)
        self.name = temp_file.name

    def delete(self):
        os.remove(self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, type, value, traceback):
        self.delete()

# Some mobile browsers which look like desktop browsers.
RE_MOBILE = re.compile(r"(iphone|ipod|blackberry|android|palm|windows\s+ce)", re.I)
RE_DESKTOP = re.compile(r"(windows|linux|os\s+[x9]|solaris|bsd)", re.I)
RE_BOT = re.compile(r"(spider|crawl|slurp|bot)", re.I)
RE_SOCIAL = re.compile(r"(facebookexternalhit/[0-9]|LinkedInBot|Twitterbot|Pinterest|Google.*snippet)", re.I)
RE_SCRAPERS = re.compile(r"(FlipboardProxy|Slurp|PaperLiBot|TweetmemeBot|MetaURI|Embedly)", re.I)

RE_EMAIL = re.compile(r"[^@]+@[^@]+\.[^@]+", re.I)

DEBUG_DATETIME = False
if settings.DEBUG_DATETIME:
    DEBUG_DATETIME = settings.DEBUG_DATETIME


def graphBuilderInplace(part, field, graph):
    output = []
    if part not in graph:
        return output
    graph_part = graph[part]
    for f in graph_part:
        if type(f) is tuple:
            f1, f2 = f
            output.append(("{0}.{1}", "{2}").format(field, f1, f2))
        else:
            output.append("{0}.{1}".format(field, f))
    return output


def graphBuilder(root_graph, field, graph):
    for part in ["fields", "recurse_into"]:
        if part not in graph:
            continue
        graph_part = graph[part]
        if part not in root_graph:
            root_graph[part] = []
        root_part = root_graph[part]
        for f in graph_part:
            if type(f) is tuple:
                f1, f2 = f
                root_part.append(("{0}.{1}".format(field, f1), f2))
            else:
                root_part.append("{0}.{1}".format(field, f))
    return root_graph

def fuzzyMatch(a, b):
    if fuzz:
        if a and b:
            a = a.lower()
            b = b.lower()
            return max(fuzz.token_set_ratio(a,b),fuzz.partial_ratio(a,b))
        return 0
    print("MISSING FUZZWUZZY MODULE")
    return 100

def isValidEmail(email):
    return bool(RE_EMAIL.search(email))

def getProtocol(request):
    if request.is_secure():
        return "https://"
    return "http://"

def getSocialReferer(request):
    referer = getReferer(request)
    if referer and "://" in referer:
        r = referer.split("/")
        domain = r[2]
        if domain in ["t.co", "twitter.com"]:
            return "twitter"
        elif domain in ["facebook.com", "m.facebook.com"]:
            return "facebook"
        elif domain in ["linkedin.com", "lnkd.in"]:
            return "linkedin"
        elif domain in ["pinterest.com"]:
            return "pinterest"
        elif domain in ["plus.google.com", "plus.url.google.com"]:
            return "googleplus"
        elif domain in ["www.google.com"]:
            return "google"
        elif domain in ["bing.com", "www.bing.com"]:
            return "bing"
        return domain
    return referer

def getReferer(request):
    return request.META.get('HTTP_REFERER')

def getRemoteIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def parseHostName(url, root=False):
    if url.startswith("http"):
        parsed_url = urlparse(url)
        hname = parsed_url.netloc
    else:
        hname = url
    if root and hname.count('.') > 1:
        return hname[hname.find(".")+1:]
    return hname


def getFromEmailForHost(host):
    email_domain = settings.EMAIL_DEFAULT_DOMAIN
    from_label = settings.EMAIL_FROM_LABEL
    if settings.EMAIL_SETTINGS:
        hostname = parseHostName(host, True)
        info = settings.EMAIL_SETTINGS.get(hostname, None)
        if info is not None:
            email_domain = info.get("email_domain", email_domain)
            from_label = info.get("from_label", from_label)
    return f"{from_label} <noreply@{email_domain}>"


def getUserAgent(request):
  # Some mobile browsers put the User-Agent in a HTTP-X header
  return request.META.get('HTTP_X_OPERAMINI_PHONE_UA') or \
         request.META.get('HTTP_X_SKYFIRE_PHONE') or \
         request.META.get('HTTP_USER_AGENT', '')

def getSocialAgent(request):
    user_agent = getUserAgent(request).lower()
    if "facebook" in user_agent:
        return "facebook"
    if "linkedin" in user_agent:
        return "linkedin"
    if "twitter" in user_agent:
        return "twitter"
    if "pinterest" in user_agent:
        return "pinterest"
    if "google" in user_agent:
        return "google"
    if "yahoo" in user_agent:
        return "yahoo"
    if "flipboard" in user_agent:
        return "flipboard"
    if "embedly" in user_agent:
        return "embedly"
    return "unknown"

def agentHas(request, keyword):
    return keyword.lower() in getUserAgent(request).lower()

def isSocialScraper(request):
    user_agent = getUserAgent(request)
    return bool(RE_SCRAPERS.search(user_agent))

def isSocialAgent(request):
    user_agent = getUserAgent(request)
    return bool(RE_SOCIAL.search(user_agent))

def isBotAgent(request):
    user_agent = getUserAgent(request)
    return bool(RE_BOT.search(user_agent))

def isMobile(request):
    user_agent = getUserAgent(request)
    return bool(RE_MOBILE.search(user_agent)) and not bool(RE_BOT.search(user_agent))

def isDesktopAgent(request):
    user_agent = getUserAgent(request)
    return not bool(RE_MOBILE.search(user_agent)) and bool(RE_DESKTOP.search(user_agent)) or bool(RE_BOT.search(user_agent))

def lowerKeys(x):
    if isinstance(x, list):
        return [lowerKeys(v) for v in x]
    elif isinstance(x, dict):
        return dict((k.lower(), lowerKeys(v)) for k, v in list(x.items()))
    return x

def mergeDicts(*args):
    context = {}
    for arg in args:
        context.update(arg)
    return context

def removeUnderscoreKeys(x):
    if isinstance(x, list):
        return [removeUnderscoreKeys(v) for v in x]
    elif isinstance(x, dict):
        return dict((k.replace('_', ''), removeUnderscoreKeys(v)) for k, v in list(x.items()))
    return x

def filterByDateRange(qset, request=None, start=None, kind=None, eod=None, field="created", end=None, zone=None):
    """
    "DateRangeStart": POSIX Start Date,
    "DateRangeEnd": POSIX End Date,
    "DateRangeEOD": Specify the UTC end of day for the range
    "DateRangeField": "created or modified",
    "DateRangeKind": "none, day, month, year",
    """
    if request:
        field = request.DATA.get(["daterangefield", "datefield"], field)
        # print(("field is: {}".format(field)))

        day = request.DATA.get("dateday", None)
        month = request.DATA.get("datemonth", None)
        year = request.DATA.get("dateyear", None)
        if day or month or year:
            # print(("running by day, month, year for field: {}".format(field)))
            qf = {}
            if day:
                qf["{}__day".format(field)] = day
            if month:
                qf["{}__month".format(field)] = month
            if year:
                qf["{}__year".format(field)] = year
            # print(qf)
            return qset.filter(**qf)

        start = request.DATA.get(["daterangestart", "date_start"], start)
        if not start:
            return qset
        start = parseDate(start)
        # print(("start is: {}".format(start)))
        zone = request.DATA.get("daterangezone", zone)
        eod = request.DATA.get("daterangeeod", eod, field_type=int)
        if eod == -1:
            zone = None
            eod = None
        elif eod is None and request.group:
            zone = request.group.timezone
            eod = request.group.getEOD(onday=start, in_local=True)
        elif eod is None:
            eod = 0
        end = request.DATA.get(["daterangeend", "date_end"], end)
        if end is None:
            end = start + timedelta(days=1)
        else:
            end = parseDate(end)

        if end <= start:
            end = start + timedelta(days=1)

        fields = request.DATA.get("daterangefields", None, field_type=list)
        if fields:
            start_field = fields[0]
            stop_field = fields[1]
            if start.hour == 0:
                # this is a hack on a straight date, lets move zone to 8 hours
                start = start - timedelta(hours=8)
            qf = {"{0}__lte".format(start_field):start, "{0}__gte".format(stop_field):start}
            return qset.filter(**qf)
        kind = request.DATA.get(["date_kind"], kind)
        # print "kind is: {}".format(kind)
        # zone = request.DATA.get("daterangezone", zone)
        # print "zone is: {}".format(zone)
        # print "eod is: {}".format(eod)
        # # print eod
        start, end = getDateRange(start=start, end=end, kind=kind, eod=eod, zone=zone)
        # print start
        # print end
    else:
        start, end = getDateRange(start=start, end=None, kind=kind, eod=eod, zone=zone)
        # print start
        # print end

    if not field:
        field = "created"

    qf = {"{0}__gte".format(field):start, "{0}__lt".format(field):end}
    return qset.filter(**qf)




def diffMinutes(t1, t2):
    diff = t1 - t2
    days, seconds = diff.days, diff.seconds
    hours = (days * 24)
    return (seconds * 60) + (hours / 60)

def diffHours(t1, t2):
    diff = t1 - t2
    days, seconds = diff.days, diff.seconds
    hours = (days * 24)
    minutes = seconds * 60
    return (minutes * 60) + hours

def getContext(request, *args, **kwargs):
    version = settings.VERSION
    if settings.DEBUG:
        version = "{0}.{1}".format(version, time.time())

    c = {
        "version":version,
        "SITE_LABEL":settings.SITE_LABEL,
        "SITE_LOGO":settings.SITE_LOGO,
        "SERVER_NAME":settings.SERVER_NAME,
        "BASE_URL":settings.BASE_URL
    }

    if request:
        c["protocol"] = getProtocol(request)
        c["request"] = request

    for k, v in kwargs.items():
        c[k] = v
    return c

def getAppNames():
    from django.apps import apps
    return [app_config.name for app_config in apps.get_app_configs()]

def getAllModels(app_name=None):
    from django.apps import apps
    if not app_name:
        return apps.get_models()
    return apps.get_app_config(app_name).get_models()


def find_nth(haystack, needle, n):
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start+len(needle))
        n -= 1
    return start


def getSetting(key, default=None):
    return settings.get(key, default)


def filterByDates(qset, start=None, end=None, date_field="created"):
    q = {}
    if start:
        q["{}__gte".format(date_field)] = start

    if end:
        q["{}__lte".format(date_field)] = end
    if q:
        return qset.filter(**q)
    return qset

def getAverage(qset, field_name):
    res = qset.aggregate(avg_result=Avg(field_name))
    if "avg_result" in res and res["avg_result"] != None:
        return res["avg_result"]
    return 0.0

def getMin(qset, field_name):
    res = qset.aggregate(min_result=Min(field_name))
    if "min_result" in res and res["min_result"] != None:
        return res["min_result"]
    return 0.0

def getMax(qset, field_name):
    res = qset.aggregate(max_result=Max(field_name))
    if "max_result" in res and res["max_result"] != None:
        return res["max_result"]
    return 0.0

def getSum(qset, *args):
    params = {}
    for field_name in args:
        key = "sum_{}".format(field_name)
        params[key] = Sum(field_name)
    # print params
    res = qset.aggregate(**params)
    results = objict()
    for field_name in args:
        key = "sum_{}".format(field_name)
        value = res.get(key, 0)
        if value is None:
            value = 0
        results[field_name] = value
    if len(args) == 1:
        return list(results.values())[0]
    return results


def countOccurences(qset, field_name):
    output = objict()
    for item in list(qset.values(field_name).annotate(count=Count(field_name))):
        output[item[field_name]] = item["count"]
    return output


def filterOR(qset, *args):
    # filterOR(qset, dict(key1=val1), dict(key2=val2))
    q = Q(**args[0])
    for arg in args[1:]:
        q = q | Q(**arg)
    if not isinstance(qset, QuerySet):
        return qset.objects.filter(q)
    return qset.filter(q)

def updateModelFromDict(model, data):
    for key in data:
        setattr(model, key, data[key])
    return model

_KEY_NOTFOUND = object()

def getValueForKeys(data, key, default=None):
    # helper method for getting first key value from list
    if type(key) is list:
        for k in key:
            v = getValueForKeys(data, k, _KEY_NOTFOUND)
            if v != _KEY_NOTFOUND:
                return v
        return default

    if "." in key:
        keys = key.split('.')
        for k in keys:
            data = getValueForKeys(data, k, _KEY_NOTFOUND)
            if data is _KEY_NOTFOUND:
                return default
        return data

    if data is None:
        return None
    return data.get(key, default)

def dictToString(d, no_truncate=False):
    from io import StringIO
    output = StringIO()
    prettyWrite(d, output, no_truncate=no_truncate)
    out = output.getvalue()
    output.close()
    return out

def prettyPrint(d, f=sys.stdout, indent=4, banner=None):
    return prettyWrite(d, f, indent, banner)


try:
    import phonenumbers
except:
    phonenumbers = None

def isPhone(value):
    try:
        int(value.replace('-', '').replace('.', '').replace(' ', ''))
        return True
    except:
        pass
    return False

def normalizePhone(value):
    if phonenumbers:
        try:
            x = phonenumbers.parse(value, "US")
        except:
            print(value)
            x = phonenumbers.parse(value, None)
        return phonenumbers.format_number(x, phonenumbers.PhoneNumberFormat.INTERNATIONAL)
    return value.replace(' ', '').replace('.', '-').replace('(', '').replace(')', '-')


from decimal import Decimal
PRETTY_INDENT = 2
PRETTY_MAX_VALUE_LENGTH = 200
PRETTY_MAX_LINES = 400
PRETTY_MAX_LENGTH = 9000

def prettyWrite(d, f=None, indent=PRETTY_INDENT, banner=None, line_count=0, no_truncate=False):
    std_output = False
    if f is None:
        std_output = True
        f = StringIO()

    prev = None
    if banner:
        f.write('---- BEGIN {} ----\n'.format(banner))
    if type(d) is list:
        prev = False
        f.write('[')
        for i in d:
            if prev:
                line_count += 1
                f.write(',\n')
            else:
                line_count += 1
                f.write('\n')

            pos = 0
            if hasattr(f, "len"):
                pos = f.len

            if not no_truncate and (line_count > PRETTY_MAX_LINES or pos >= PRETTY_MAX_LENGTH):
                f.write('{}"...truncated"'.format(' ' * indent))
                break
            prev = True
            if type(i) is bool:
                i = int(i)
            if type(i) in [str, str]:
                if not no_truncate and len(i) >= PRETTY_MAX_VALUE_LENGTH:
                    f.write('{}"{}...truncated"'.format(' ' * indent, i[:PRETTY_MAX_VALUE_LENGTH-20]))
                else:
                    f.write('{}"{}"'.format(' ' * indent, i))
            elif type(i) is list or isinstance(i, dict):
                f.write(' ' * (indent))
                line_count = prettyWrite(i, f, indent+PRETTY_INDENT, line_count=line_count)
            elif type(i) is Decimal:
                f.write('{}{}'.format(' ' * indent, str(i)))
            else:
                f.write('{}"{}"'.format(' ' * indent, i))
        line_count += 1
        f.write('\n')
        f.write(' ' * (indent-PRETTY_INDENT))
        f.write(']')
    elif isinstance(d, dict):
        f.write('{')
        for key, value in list(d.items()):
            if prev:
                line_count += 1
                f.write(',\n')
            else:
                line_count += 1
                f.write('\n')

            pos = 0
            if hasattr(f, "len"):
                pos = f.len

            depth = indent / PRETTY_INDENT
            if not no_truncate and  (depth > 1 and (line_count > PRETTY_MAX_LINES or pos >= PRETTY_MAX_LENGTH)):
                f.write('{}"truncated":"...truncated"\n'.format(' ' * indent))
                break
            prev = True
            if type(key) in [str, str]:
                f.write('{}"{}":'.format(' ' * indent, key))
            else:
                f.write('{}{}: '.format(' ' * indent, str(key)))
            if type(value) is list or isinstance(value, dict):
                f.write(' ')
                line_count = prettyWrite(value, f, indent+PRETTY_INDENT, line_count=line_count)
            else:
                if type(value) in [str, str]:
                    if not no_truncate and len(value) >= PRETTY_MAX_VALUE_LENGTH:
                        f.write(' "{}...truncated"'.format(value[:PRETTY_MAX_VALUE_LENGTH-20]))
                    else:
                        f.write(' "{}"'.format(value))
                elif type(value) in [datetime, time]:
                    f.write(' "{}"'.format(value))
                elif type(value) is Decimal:
                    f.write(' {}'.format(str(value)))
                else:
                    f.write(' {}'.format(value))
        line_count += 1
        f.write('\n')
        f.write(' ' * (indent-PRETTY_INDENT))
        f.write('}')
    else:
        f.write(str(d))

    if banner:
        f.write('\n---- END {} ----\n'.format(banner))
    if indent == PRETTY_INDENT:
        f.write("\n")
    if std_output:
        sys.stdout.write(f.getvalue())
        f.close()
    return line_count

def randomKey(count=6):
    return ''.join(random.choice(string.uppercase + string.digits) for x in range(count))

def getStackString():
    import traceback
    return str(traceback.format_exc())

def isPidRunning(pid):
    if sys.platform.startswith("linux"):
        return os.path.exists("/proc/{}".format(pid))
    try:
        os.kill(pid, 0)  # or signal.SIGKILL
    except OSError as err:
        return False
    return True

def getPidFromFile(pid_file):
    pid = 0
    if os.path.exists(pid_file):
        with open(pid_file, "r") as f:
            pid = f.read()
            if pid.isdigit():
                pid = int(pid)
    return pid


def sudoCMD(cmd, as_user=None, test_first=True):
    if as_user and as_user == getUsername():
        sudo_cmd = []
        if isinstance(cmd, list):
            sudo_cmd.extend(cmd)
        else:
            sudo_cmd.append(cmd)
        return subprocess.Popen(sudo_cmd, close_fds=True)
    test_sudo_cmd = ["sudo", "-lU"]
    # -n for non-interactive, will print error if password required
    sudo_cmd = ["sudo", "-n"]
    if as_user:
        test_sudo_cmd.append("-lU")
        test_sudo_cmd.append(as_user)
        sudo_cmd.append("-u")
        sudo_cmd.append(as_user)
    else:
        test_sudo_cmd.append("root")
    if isinstance(cmd, list):
        test_sudo_cmd.extend(cmd)
        sudo_cmd.extend(cmd)
    else:
        test_sudo_cmd.append(cmd)
        sudo_cmd.append(cmd)
    if test_first:
        print(test_sudo_cmd)
        process = subprocess.Popen(test_sudo_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        if err.strip():
            raise Exception("WARNING: cannot run {}, we don't have sudo rights".format(test_sudo_cmd))
    return subprocess.Popen(sudo_cmd, close_fds=True)


def blockIP(ip):
    # for global Task.Publish("incident", "global_block", {ip:"x.x.x.x"}, channel="tq_broadcast")
    log_error(f"blocking {ip}")
    return sudoCMD([settings.BLOCK_IP_CMD, "add", ip])


def unblockIP(ip):
    # for global Task.Publish("incident", "global_unblock", {ip:"x.x.x.x"}, channel="tq_broadcast")
    return sudoCMD([settings.BLOCK_IP_CMD, "delete", ip])


def getBlockedIPs():
    return sudoCMD([settings.BLOCK_IP_CMD, "list"])


def getBlockedHosts():
    blocked = []
    if not os.path.exists("/etc/hosts.deny"):
        return blocked
    with open("/etc/hosts.deny", 'r') as f:
        for line in f.readlines():
            if line.startswith("#"):
                continue
            if line.startswith("ALL:"):
                line = line.split("L:")[1]
            blocked.append(line.strip())
    return blocked


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value

def toBytes(value):
    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, bytearray):
        value = bytes(value)
    return value

def toByteArray(value):
    if isinstance(value, bytearray):
        return value
    elif isinstance(value, str):
        return bytearray(value, 'utf-8')
    elif isinstance(value, bytes):
        return bytearray(value)
    return value

def toHex(value):
    return toBytes(value).hex().upper()

def hexToByteArray(value):
    return bytearray.fromhex(value)

def hexToString(value):
    return bytes.fromhex(value).decode('utf-8')

def toBase64(value):
    return base64.b64encode(value)

def fromBase64(value):
    return toString(base64.b64decode(value))

def removeNonAscii(input_str):
    """Remove all non-ASCII characters from the input string."""
    return ''.join(char for char in input_str if 0 <= ord(char) < 128)


def dictDeepUpdate(dict1, dict2):
    for key, value in dict2.items():
        if isinstance(value, dict):
            if key not in dict1:
                dict1[key] = objict()
            dictDeepUpdate(dict1[key], value)
        else:
            dict1[key] = dict2[key]
    return dict1
    



import http.client
from . import httplib2
import mimetypes

from medialib.youtube.apiclient.errors import HttpError
from medialib.youtube.apiclient.http import MediaFileUpload

import random
import time

# Maximum number of times to retry before giving up.
MAX_RETRIES = 10

# Always retry when these exceptions are raised.
RETRIABLE_EXCEPTIONS = (httplib2.HttpLib2Error, IOError, http.client.NotConnected,
  http.client.IncompleteRead, http.client.ImproperConnectionState,
  http.client.CannotSendRequest, http.client.CannotSendHeader,
  http.client.ResponseNotReady, http.client.BadStatusLine)

# Always retry when an apiclient.errors.HttpError with one of these status
# codes is raised.
RETRIABLE_STATUS_CODES = [500, 502, 503, 504]

def resumable_upload(service, videofile, title, description, category="22", keywords="", privacy="public"):
    """
    Resumable upload for video files.

    returns

    """
    tags = None
    if keywords:
        tags = keywords.split(",")

    body=dict(
        snippet=dict(
            title=title,
            description=description,
            tags=tags,
            categoryId=category
        ),
        status=dict(
            privacyStatus=privacy
        )
    )

    (mimetype, encoding) = mimetypes.guess_type(videofile)
    if mimetype is None and (".uploadedMovie" in videofile or ".m4v" in videofile):
        mimetype = "video/quicktime"

    if mimetype is None:
        raise Exception("unable to determine mimetype of file")
    # Call the API's videos.insert method to create and upload the video.
    insert_request = service.videos().insert(
        part=",".join(list(body.keys())),
        body=body,
        media_body=MediaFileUpload(videofile, mimetype=mimetype, chunksize=-1, resumable=True)
    )

    response = None
    error = None
    retry = 0
    while response is None:
        try:
            status, response = insert_request.next_chunk()
            if 'id' in response:
                service.logger.debug("Video id '%s' was successfully uploaded." % response['id'])
                return response
            else:
                raise Exception("The upload failed with an unexpected response: %s" % response)
        except HttpError as e:
            if e.resp.status in RETRIABLE_STATUS_CODES:
                error = "A retriable HTTP error %d occurred:\n%s" % (e.resp.status, e.content)
            else:
                raise
        except RETRIABLE_EXCEPTIONS as e:
            error = "A retriable error occurred: %s" % e

        if error is not None:
            service.logger.error(error)
            retry += 1
            if retry > MAX_RETRIES:
                raise Exception("No longer attempting to retry.")

            max_sleep = 2 ** retry
            sleep_seconds = random.random() * max_sleep
            time.sleep(sleep_seconds)



# Generated by Django 4.1.1 on 2022-09-24 21:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0009_alter_member_phone_number'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AuthToken',
        ),
    ]



# Generated by Django 4.1.4 on 2023-05-04 20:11

from django.db import migrations, models
import rest.models


class Migration(migrations.Migration):

    dependencies = [
        ('metrics', '0003_metrics_expires'),
    ]

    operations = [
        migrations.CreateModel(
            name='EODMetrics',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('end', models.DateTimeField(db_index=True)),
                ('uuid', models.SlugField(max_length=250, unique=True)),
                ('slug', models.CharField(db_index=True, default=None, max_length=200, null=True)),
                ('expires', models.DateTimeField(db_index=True, default=None, null=True)),
                ('granularity', models.CharField(db_index=True, max_length=64)),
                ('component', models.CharField(db_index=True, default=None, max_length=255, null=True)),
                ('component_pk', models.IntegerField(db_index=True, default=None, null=True)),
                ('group_id', models.IntegerField(db_index=True, default=None, null=True)),
                ('k1', models.SlugField(default=None, max_length=64, null=True)),
                ('v1', models.IntegerField(db_index=True, default=0)),
                ('k2', models.SlugField(default=None, max_length=64, null=True)),
                ('v2', models.IntegerField(db_index=True, default=0)),
                ('k3', models.SlugField(default=None, max_length=64, null=True)),
                ('v3', models.IntegerField(default=0)),
                ('k4', models.SlugField(default=None, max_length=64, null=True)),
                ('v4', models.IntegerField(default=0)),
                ('k5', models.SlugField(default=None, max_length=64, null=True)),
                ('v5', models.IntegerField(default=0)),
                ('k6', models.SlugField(default=None, max_length=64, null=True)),
                ('v6', models.IntegerField(default=0)),
                ('k7', models.SlugField(default=None, max_length=64, null=True)),
                ('v7', models.IntegerField(default=0)),
                ('k8', models.SlugField(default=None, max_length=64, null=True)),
                ('v8', models.IntegerField(default=0)),
                ('k9', models.SlugField(default=None, max_length=64, null=True)),
                ('v9', models.IntegerField(default=0)),
                ('k10', models.SlugField(default=None, max_length=64, null=True)),
                ('v10', models.IntegerField(default=0)),
                ('k11', models.SlugField(default=None, max_length=64, null=True)),
                ('v11', models.IntegerField(default=0)),
                ('k12', models.SlugField(default=None, max_length=64, null=True)),
                ('v12', models.IntegerField(default=0)),
                ('k13', models.SlugField(default=None, max_length=64, null=True)),
                ('v13', models.IntegerField(default=0)),
                ('k14', models.SlugField(default=None, max_length=64, null=True)),
                ('v14', models.IntegerField(default=0)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



from __future__ import unicode_literals
from datetime import datetime, timedelta

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.core.management.base import NoArgsCommand
from django.template.loader import render_to_string

from redis_metrics.utils import get_r


class Command(NoArgsCommand):
    help = "Send Metrics Report E-mails"
    can_import_settings = True

    def handle_noargs(self, **options):
        """Send Report E-mails."""

        r = get_r()
        since = datetime.utcnow() - timedelta(days=1)
        metrics = {}
        categories = r.metric_slugs_by_category()
        for category_name, slug_list in categories.items():
            metrics[category_name] = []
            for slug in slug_list:
                metric_values = r.get_metric_history(slug, since=since)
                metrics[category_name].append(
                    (slug, metric_values)
                )

        # metrics is now:
        # --------------
        # { Category : [
        #     ('foo', [('m:foo:2012-07-18', 1), ('m:foo:2012-07-19, 2), ...])
        #   ],
        #   ...
        # }

        template = "redis_metrics/email/report.{fmt}"
        data = {
            'today': since,
            'metrics': metrics,
        }

        message = render_to_string(template.format(fmt='txt'), data)
        message_html = render_to_string(template.format(fmt='html'), data)
        msg = EmailMultiAlternatives(
            subject="Redis Metrics Report",
            body=message,
            from_email=settings.DEFAULT_FROM_EMAIL,
            to=[email for name, email in settings.ADMINS]
        )
        msg.attach_alternative(message_html, "text/html")
        msg.send()



from . import models as auditlog

from rest import views as rv
from rest import decorators as rd


@rd.url(r'^plog$')
@rd.url(r'^plog/(?P<plog_id>\d+)$')
@rd.perm_required('view_logs')
def plog_handler(request, plog_id=None):
    if not plog_id:
        min_pk = getattr(auditlog.settings, "PLOG_STALE_ID", 0)
        return auditlog.PersistentLog.on_rest_list(request, qset=auditlog.PersistentLog.objects.filter(pk__gte=min_pk))
    return auditlog.PersistentLog.on_rest_request(request, plog_id)


@rd.urlGET(r'^plog_old$')
@rd.perm_required('view_logs')
def plogList(request):
    auditlog.PersistentLog.on_request_handle()

    graph = request.DATA.get("graph", "default")
    qset = auditlog.PersistentLog.objects.all()
    if request.group:
        qset = qset.filter(group=request.group)

    ip = request.DATA.get("ip")
    if ip:
        qset = qset.filter(session__ip=ip)

    path = request.DATA.get("path")
    if path:
        qset = qset.filter(remote_path__icontains=path)
    
    method = request.DATA.get("method")
    if method:
        qset = qset.filter(method=method)

    action = request.DATA.get("action")
    if action:
        qset = qset.filter(action=action)
    
    component = request.DATA.get("component")
    if component:
        qset = qset.filter(component=component)
    
    pkey = request.DATA.get("pkey")
    if pkey:
        qset = qset.filter(pkey=pkey)

    username = request.DATA.get("username")
    if username:
        qset = qset.filter(user__username=username)
    
    term = request.DATA.get("term")
    if term:
        qset = qset.filter(message__icontains=term)
    
    return rv.restList(request, qset.order_by('-when'), **auditlog.PersistentLog.getGraph(graph))


@rd.urlGET (r'^list$')
def auditList(request):
    """
    | Parameter: model=<string> (default=all)
    | Parameter: id=<int> (default=all)
    | Parameter: attributes=<string> (default=all)
    | Parameter: user=<id|username> (default=all)
    |
    | Return: audit log entries
    """
    ret = auditlog.AuditLog.objects.all()
    
    model = request.DATA.get('model')
    if model:
        ret = ret.filter(model=model)

    id = request.DATA.get('id')
    if id:
        ret = ret.filter(pkey=int(id))

    attributes = request.DATA.get('attributes')
    if attributes:
        attributes = attributes.split(',')
        print(attributes)
        ret = ret.filter(attribute__in=attributes)

    user = request.DATA.get('user')
    if user:
        ret = ret.filter(user=user)

    return rv.restList(
        request, ret, sort='-when',
        fields=(
            'when',
            'model',
            'pkey',
            'attribute',
            'user.id',
            'user.username',
            'reference',
            'how',
            'referer',
            'stack',
            'oldval',
            'newval',
        ), recurse_into=(
            'user',
        ), require_perms=(
            'auditlog.can_read',
        ),
    )


@rd.url(r'^nuke$')
@rd.perm_required("nuke_logs")
def nuke_log_data(request):
    # nuclear option used for testing system to clear out all wallet data
    # make sure this is enabled for this setup
    confirm_nuke = request.DATA.get("nuke_code", None)
    if confirm_nuke != "launch" or not getattr(auditlog.settings, "CAN_NUKE_DATABASE", False):
        return rv.restPermissionDenied(request)

    # first nuke all transactions
    auditlog.PersistentLog.objects.all().delete()
    auditlog.AuditLog.objects.all().delete()
    auditlog.SessionLog.objects.all().delete()
    return rv.restStatus(request, True)



from django.template.loader import render_to_string
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from rest import settings
import os
from objict import objict
from inbox.models.template import MailTemplate
import csv
from io import StringIO
import re


def createMessage(sender, recipients, subject, text, html, attachments=None, replyto=None):
    message = objict(sender=sender, recipients=recipients)

    if not isinstance(message.recipients, (tuple, list)):
        message.recipients = [message.recipients]

    if attachments is None:
        attachments = []

    message.msg = createMultiPartMessage(
        message.sender, message.recipients, subject,
        text=text, html=html, attachments=attachments,
        replyto=replyto)
    return message


def createMultiPartMessage(sender, recipients, subject, text, html, attachments, replyto):
    """
    Creates a MIME multipart message object.
    Uses only the Python `email` standard library.
    Emails, both sender and recipients, can be just the email string or have the format 'The Name <the_email@host.com>'.

    :param sender: The sender.
    :param recipients: List of recipients. Needs to be a list, even if only one recipient.
    :param subject: The subject of the email.
    :param text: The text version of the email body (optional).
    :param html: The html version of the email body (optional).
    :param attachments: List of files to attach in the email.
    :param replyto: optional reply to address.
    :return: A `MIMEMultipart` to be used to send the email.
    """
    multipart_content_subtype = 'alternative' if text and html else 'mixed'
    msg = MIMEMultipart(multipart_content_subtype)
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = ', '.join(recipients)
    if replyto:
        msg.add_header('reply-to', replyto)

    # Record the MIME types of both parts - text/plain and text/html.
    # According to RFC 2046, the last part of a multipart message, in this case the HTML message, is best and preferred.
    if text:
        part = MIMEText(text, 'plain')
        msg.attach(part)
    if html:
        # HACK to remove codec errors
        html = html.encode('ascii', 'ignore').decode('ascii')
        part = MIMEText(html, 'html')
        msg.attach(part)

    # Add attachments
    index = 0
    for atch in attachments or []:
        if isinstance(atch, (str, bytes)):
            index += 1
            atch = objict(name="attachment{}.txt".format(index), data=atch, mimetype="text/plain")
        # lets attach it
        part = MIMEApplication(atch.data)
        part.add_header('Content-Type', atch.mimetype)
        part.add_header('Content-Disposition', 'attachment', filename=os.path.basename(atch.name))
        msg.attach(part)
    return msg


def renderTemplate(template, context, group=None):
    context["SITE_LABEL"] = settings.SITE_LABEL
    context["BASE_URL"] = settings.BASE_URL
    context["SITE_LOGO"] = settings.SITE_LOGO
    context["SERVER_NAME"] = settings.SERVER_NAME
    context["version"] = settings.VERSION

    if template[-4:] in ["html", ".txt"]:
        # this is a django template
        return render_to_string(template, context)
    qset = MailTemplate.objects.filter(name=template)
    if group is not None:
        gqset = qset.filter(group=group)
        mtemp = gqset.last()
        if mtemp is None:
            mtemp = qset.last()
    else:
        mtemp = qset.last()
    if mtemp is None:
        return None
    return mtemp.render(context)


def generateCSV(qset, fields, name):
    a = objict()
    a.name = name
    a.file = StringIO.StringIO()
    csvwriter = csv.writer(a.file)
    csvwriter.writerow(fields)

    for row in qset.values_list(*fields):
        row = [str(x) for x in row]
        csvwriter.writerow(row)
    a.data = a.file.getvalue()
    a.mimetype = "text/csv"
    return a


def isHTML(text):
    pattern = r'<[a-zA-Z0-9]+>.*?<\/[a-zA-Z0-9]+>'
    if re.search(pattern, text):
        return True
    return False



from __future__ import unicode_literals
import re

from datetime import date
from django.core.management.base import BaseCommand, CommandError
from redis_metrics.utils import r


class Command(BaseCommand):
    """

    Updates the weekly metrics so they use the new key format. For
    more information, see Issue #7: http://bit.ly/YEjtF9

    Essentially, weekly-aggregated metrics had a key of the form:

        m:<slug>:w:<nn>

    Where ``nn`` was the week number in the range: [00-52). They now take the
    form:

        m:<slug>:w:<yyyy-nn>

    Where ``yyyy`` is a 4-digit year, and ``nn`` is the week number.

    """
    args = '[year]'
    help = "Updates weekly metrics so they match the new key format"

    def handle(self, *args, **options):
        if len(args) == 0:
            # Use the current year
            year = date.today().year
        elif len(args) == 1:
            year = int(args[0])
        else:
            raise CommandError("Invalid arguments. Provide a 4-digit year.")

        r = R()

        # Retrieve all the metric keys of the form: "m:<slug>:w:<nn>"
        weekly_keys = filter(
            lambda k: re.match(r'^m:.+:w:\d\d', k) is not None,
            r.r.smembers(r._metric_slugs_key)
        )
        for old_key in weekly_keys:
            # Match   -> m:<slug>:w:<nn>
            # Replace -> m:<slug>:w:<yyyy-nn>
            parts = old_key.split(":")
            parts[-1] = "{0}-{1}".format(year, parts[-1])
            new_key = ":".join(parts)

            # Copy old metric value into new metric
            r.r.set(new_key, r.r.get(old_key))

            # Track the new key
            r.r.sadd(r._metric_slugs_key, new_key)

            # Delete the old metric?
            r.r.delete(old_key)

        # Delete the set of keys for the old weekly metrics
        r.r.srem(r._metric_slugs_key, *weekly_keys)




import os
import threading
from io import StringIO
import csv
import mimetypes
import boto3

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication

try:
    from inlinestyler.utils import inline_css
except Exception:
    inline_css = None

from django.template.loader import render_to_string
from django.template import TemplateDoesNotExist

import metrics
from rest import settings
from rest.uberdict import UberDict
from rest.middleware import get_request
from rest.log import getLogger

EMAIL_LOGGER = getLogger("email", filename="email.log")
# ses settings
SES_ACCESS_KEY = settings.SES_ACCESS_KEY
SES_SECRET_KEY = settings.SES_SECRET_KEY
SES_REGION = settings.SES_REGION
EMAIL_METRICS = settings.get("EMAIL_METRICS", False)


def send(to, subject, body=None, attachments=[], 
         from_email=settings.DEFAULT_FROM_EMAIL,
         fail_silently=True, template=None, 
         context=None, do_async=False, replyto=None):
    # make sure to is list
    if not isinstance(to, (tuple, list)):
        to = [to]
    # if template lets render
    if isinstance(body, (tuple, list)):
        text, html = body
    else:
        text = None
        html = body

    if template:
        html = renderBody(html, template, context)

    msg = createMultiPartMessage(
        from_email, to, subject,
        text=text, html=html, attachments=attachments,
        replyto=replyto)

    # send now or async
    if not do_async:
        return sendMail(msg, from_email, to, fail_silently=fail_silently)
    else:
        t = threading.Thread(target=sendMail, args=[msg, from_email, to])
        t.start()
    return True


def renderBody(body, template=None, context=None):
    if template and body and not context:
        context = {
            "body": body
        }

    if template and context:
        if isinstance(context, dict):
            context['settings'] = settings
        if template[-4:] not in ["html", ".txt"]:
            template += ".html"
        body = render_to_string(template, context)
        if inline_css:
            body = inline_css(body)
    return body


def createMultiPartMessage(sender, recipients, subject, text, html, attachments, replyto):
    """
    Creates a MIME multipart message object.
    Uses only the Python `email` standard library.
    Emails, both sender and recipients, can be just the email string or have the format 'The Name <the_email@host.com>'.

    :param sender: The sender.
    :param recipients: List of recipients. Needs to be a list, even if only one recipient.
    :param subject: The subject of the email.
    :param text: The text version of the email body (optional).
    :param html: The html version of the email body (optional).
    :param attachments: List of files to attach in the email.
    :param replyto: optional reply to address.
    :return: A `MIMEMultipart` to be used to send the email.
    """
    multipart_content_subtype = 'alternative' if text and html else 'mixed'
    msg = MIMEMultipart(multipart_content_subtype)
    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = ', '.join(recipients)
    if replyto:
        msg.add_header('reply-to', replyto)

    # Record the MIME types of both parts - text/plain and text/html.
    # According to RFC 2046, the last part of a multipart message, in this case the HTML message, is best and preferred.
    if text:
        part = MIMEText(text, 'plain')
        msg.attach(part)
    if html:
        # HACK to remove codec errors
        html = html.encode('ascii', 'ignore').decode('ascii')
        part = MIMEText(html, 'html')
        msg.attach(part)

    # Add attachments
    index = 0
    for atch in attachments or []:
        if isinstance(atch, (str, bytes)):
            index += 1
            atch = UberDict(name="attachment{}.txt".format(index), data=atch, mimetype="text/plain")
        # lets attach it
        part = MIMEApplication(atch.data)
        part.add_header('Content-Type', atch.mimetype)
        part.add_header('Content-Disposition', 'attachment', filename=os.path.basename(atch.name))
        msg.attach(part)
    return msg


def sendMail(msg, sender, recipients, fail_silently=True):
    try:
        ses_client = getSES(SES_ACCESS_KEY, SES_SECRET_KEY, SES_REGION)
        ses_client.send_raw_email(
            Source=sender,
            Destinations=recipients,
            RawMessage={'Data': msg.as_string()}
        )
        if EMAIL_METRICS:
            metrics.metric("emails_sent", category="email", min_granularity="hourly")
        return True
    except Exception as err:
        if EMAIL_METRICS:
            metrics.metric("email_errors", category="email", min_granularity="hourly")
        EMAIL_LOGGER.exception(err)
        EMAIL_LOGGER.error(msg.as_string())
        if not fail_silently:
            raise err
    return False


def makeAttachment(filename, data):
    atment = UberDict(name=filename, data=data)
    atment.mimetype, junk = mimetypes.MimeTypes().guess_type(filename)
    return atment


def getSES(access_key, secret_key, region):
    return boto3.client(
        'ses',
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        region_name=region)


def generateCSV(qset, fields, name):
    a = UberDict()
    a.name = name
    a.file = StringIO.StringIO()
    csvwriter = csv.writer(a.file)
    csvwriter.writerow(fields)

    for row in qset.values_list(*fields):
        row = [str(x) for x in row]
        csvwriter.writerow(row)
    a.data = a.file.getvalue()
    a.mimetype = "text/csv"
    return a


def render_to_mail(name, context):
    if not context.get("request"):
        context["request"] = get_request()
    context['newline'] = "\n"
    toaddrs = None
    if 'to' in context:
        toaddrs = context['to']
        if type(toaddrs) != list:
            toaddrs = [toaddrs]
    else:
        try:
            toaddrs = render_to_string(name + ".to", context).splitlines()
        except TemplateDoesNotExist:
            return
    try:
        while True:
            toaddrs.remove('')
    except ValueError:
        pass
    if len(toaddrs) == 0:
        EMAIL_LOGGER.error("Sending email to no one: {}".format(name), context)
        return

    try:
        html_content = render_to_string(name + ".html", context)
        if inline_css:
            html_content = inline_css(html_content)
    except TemplateDoesNotExist:
        html_content = None
        pass

    text_content = ""
    try:
        text_content = render_to_string(name + ".txt", context)
    except TemplateDoesNotExist:
        if html_content is not None:
            raise TemplateDoesNotExist("requires at least one content template")

    if 'from' in context:
        fromaddr = context['from']
    else:
        try:
            fromaddr = render_to_string(name + ".from", context).rstrip()
        except TemplateDoesNotExist:
            fromaddr = settings.DEFAULT_FROM_EMAIL
    # print fromaddr
    if 'subject' in context:
        subject = context['subject']
    else:
        try:
            subject = render_to_string(name + ".subject", context).rstrip()
        except TemplateDoesNotExist:
            EMAIL_LOGGER.error("Sending email without subject: %s" % name)
            return False
    replyto = None
    if "replyto" in context:
        replyto = context["replyto"]

    body = None
    if bool(text_content) and bool(html_content):
        body = (text_content, html_content)
    elif bool(text_content):
        body = text_content
    elif bool(html_content):
        body = html_content

    send(toaddrs, subject, body=body, from_email=fromaddr, do_async=True, replyto=replyto)





from rest import decorators as rd
from rest import views as rv
from pushit.models import Product, Release


@rd.url('product')
@rd.url('product/<int:pk>')
def rest_on_product(request, pk=None):
    return Product.on_rest_request(request, pk)


@rd.url('release')
@rd.url('release/<int:pk>')
def rest_on_release(request, pk=None):
    return Release.on_rest_request(request, pk)



from . import s3store
from . import httpstore
from . import youtubestore
from . import filestore
from . import rtmpstore
from . import nullstore

from django.conf import settings

try:
    import boto.cloudfront
except:
    pass

from urllib.parse import urlparse
import time
from django.core.cache import cache

def _pick_store(url):
    if url[:3] == "s3:":
        return s3store
    elif url[:5] == "http:" or url[:6] == "https:":
        return httpstore
    elif url[:5] == "file:":
        return filestore
    elif url[:8] == "youtube:":
        return youtubestore
    elif url[:5] == "rtmp:":
        return rtmpstore
    elif url[:1] == "/":
        return nullstore
    elif url == "":
        return nullstore
    else:
        raise Exception("Unknown URL format: %s" % url)

def type(url):
    store = _pick_store(url)
    return store.__name__.split(".")[-1]

def upload(url, fp, background=False):
    store = _pick_store(url)
    return store.upload(url, fp, background)

def view_url(url, expires=600, is_secure=False, request=None):
    mirrors = getattr(settings, 'CLOUDFRONT_MIRRORS', {})
    for k in mirrors:
        if url[:len(k)] == k:
            if is_secure and "url_secure" in mirrors[k]:
                murl = mirrors[k]["url_secure"]
            else:
                murl = mirrors[k]["url"]
            mpath = url[len(k):]
            try:
                ext = mpath.rsplit('.', 1)[1]
            except ValueError:
                ext = None
            srcip = getattr(request, 'META', {}).get('REMOTE_ADDR', None)
            try:
                ips = list(int(a) for a in srcip.split('.'))
                if len(ips) == 4 and (
                    (ips[0] == 127) or
                    (ips[0] == 10) or
                    (ips[0] == 192 and ips[1] == 168) or
                    (ips[0] == 172 and ips[1] >= 16 and ips[1] <= 31) or
                    (ips[0] == 204 and ips[1] == 97 and ips[2] == 176)):
                    srcip = None
            except (ValueError, AttributeError,):
                srcip = None

            if not 'key' in mirrors[k] or ext in mirrors[k].get('no_sign', []):
                return murl + mpath
            ckey = "cfsign/" + (srcip or '') + "/" + murl
            sign = cache.get(ckey)
            if sign and expires != None:
                if '?' in mpath:
                    sep = '&'
                else:
                    sep = '?'
                return murl + mpath + sep + sign

            dist = boto.cloudfront.Distribution()
            purl = murl + mpath

            try:
                if expires == None:
                    surl = dist.create_signed_url(purl, mirrors[k]["key"], private_key_string=mirrors[k]["secret"], expire_time=2147483647)
                    return surl
                else:
                    surl = dist.create_signed_url(purl, mirrors[k]["key"], private_key_string=mirrors[k]["secret"],
                        expire_time=int(time.time())+expires+1800, ip_address=srcip, policy_url=murl + "*")
            except (AttributeError, NotImplementedError,):
                #print "Your installation of boto does not support cloudfront"
                break
            cache.set(ckey, surl[len(purl)+1:], 1800)
            return surl

    store = _pick_store(url)
    return store.view_url(url, expires=expires, is_secure=is_secure)

def rtmp_url(url, expires=600, is_secure=False, request=None):
    mirrors = getattr(settings, 'CLOUDFRONT_MIRRORS', {})
    for k in mirrors:
        if url[:len(k)] == k:
            if "url_rtmp" in mirrors[k]:
                murl = mirrors[k]["url_rtmp"]
            else:
                break

            mpath = url[len(k):]
            try:
                (mpath, ext) = mpath.rsplit('.', 1)
            except ValueError:
                break
            if ext in mirrors[k].get("rtmp_exts", ()):
                mpath = ext + ":" + mpath
            else:
                break

            srcip = getattr(request, 'META', {}).get('REMOTE_ADDR', None)
            try:
                ips = list(int(a) for a in srcip.split('.'))
                if len(ips) == 4 and (
                    (ips[0] == 127) or
                    (ips[0] == 10) or
                    (ips[0] == 192 and ips[1] == 168) or
                    (ips[0] == 172 and ips[1] >= 16 and ips[1] <= 31) or
                    (ips[0] == 204 and ips[1] == 97 and ips[2] == 176)):
                    srcip = None
            except (ValueError, AttributeError,):
                srcip = None

            ckey = "cfsign/" + (srcip or '') + "/"
            sign = cache.get(ckey)
            if sign:
                if '?' in mpath:
                    sep = '&'
                else:
                    sep = '?'
                return murl + mpath + sep + sign

            dist = boto.cloudfront.Distribution()
            purl = murl + mpath
            try:
                surl = dist.create_signed_url(purl, mirrors[k]["key"], private_key_string=mirrors[k]["secret"],
                    expire_time=int(time.time())+expires+1800, ip_address=srcip, policy_url="*")
            except (AttributeError, NotImplementedError,):
                #print "Your installation of boto does not support cloudfront"
                break
            cache.set(ckey, surl[len(purl)+1:], 1800)
            return surl
    return None

def exists(url):
    store = _pick_store(url)
    return store.exists(url)

def get_file(url, fp):
    store = _pick_store(url)
    return store.get_file(url, fp)

def delete(url):
    store = _pick_store(url)
    return store.delete(url)

def netloc(url):
    store = _pick_store(url)
    if hasattr(store, "netloc"):
        return store.netloc(url)
    return None

def path(url):
    store = _pick_store(url)
    if hasattr(store, "path"):
        return store.path(url)
    return None



from django.contrib.sessions.backends.base import SessionBase
from ws4redis import client as redis
from rest import settings


class SessionStore(SessionBase):
    def __init__(self, session_key=None):
        super(SessionStore, self).__init__(session_key)
        self._exists = None

    def load(self):
        try:
            sk = self._get_or_create_session_key()
            session_data = redis.get(self.get_stored_key(sk))
            return self.decode(session_data)
        except Exception:
            pass
        self._session_key = None
        return {}

    def exists(self, session_key=None):
        if session_key is None:
            session_key = self.session_key
        if session_key is None:
            return False
        if self._exists is None:
            self._exists = redis.exists(self.get_stored_key(session_key))
        return self._exists

    def create(self):
        self._session_key = self._get_new_session_key()
        self.save(True)
        self.modified = True
        return

    def save(self, must_create=False):
        if self.session_key is None:
            return self.create()
        data = self.encode(self._get_session(no_load=must_create))
        redis.set(
            self.get_stored_key(self._get_or_create_session_key()),
            data,
            self.get_expiry_age())
        return True

    def delete(self, session_key=None):
        if session_key is None:
            session_key = self.session_key
        if session_key:
            redis.delete(session_key)

    @classmethod
    def clear_expired(cls):
        pass

    def get_stored_key(self, session_key):
        """Return the real key name in redis storage
        @return string
        """
        prefix = settings.SESSION_REDIS_PREFIX
        if not prefix:
            return session_key
        return f"{prefix}{session_key}"




from urllib.parse import urlparse

def upload(url, fp, background=False):
    raise Exception("Cannot upload to youtube using store")

def view_url(url, expires=600, is_secure=False):
    u = urlparse(url)
    return 'http://www.youtube.com/watch?v=' + u.path

def exists(url):
    return None

def get_file(url, fp):
    raise Exception("Cannot download from youtube")

def delete(url):
    pass # no-op



# Generated by Django 4.0.2 on 2022-02-26 15:59

from django.db import migrations, models
import rest.fields


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0001_initial'),
        ('medialib', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='member',
            name='phone_number',
            field=rest.fields.FormattedField(default=None, max_length=64, null=True),
        ),

        migrations.AddField(
            model_name='member',
            name='picture',
            field=models.ForeignKey(blank=True, help_text='Profile picture', null=True, on_delete=models.deletion.CASCADE, related_name='+', to='medialib.mediaitem'),
        )
    ]


# Copyright (C) 2013 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module holds the old run() function which is deprecated, the
tools.run_flow() function should be used in its place."""


import logging
import socket
import sys
import webbrowser

import gflags

from oauth2client import client
from oauth2client import util
from .tools import ClientRedirectHandler
from .tools import ClientRedirectServer


FLAGS = gflags.FLAGS

gflags.DEFINE_boolean('auth_local_webserver', True,
                      ('Run a local web server to handle redirects during '
                       'OAuth authorization.'))

gflags.DEFINE_string('auth_host_name', 'localhost',
                     ('Host name to use when running a local web server to '
                      'handle redirects during OAuth authorization.'))

gflags.DEFINE_multi_int('auth_host_port', [8080, 8090],
                        ('Port to use when running a local web server to '
                         'handle redirects during OAuth authorization.'))


@util.positional(2)
def run(flow, storage, http=None):
  """Core code for a command-line application.

  The run() function is called from your application and runs through all
  the steps to obtain credentials. It takes a Flow argument and attempts to
  open an authorization server page in the user's default web browser. The
  server asks the user to grant your application access to the user's data.
  If the user grants access, the run() function returns new credentials. The
  new credentials are also stored in the Storage argument, which updates the
  file associated with the Storage object.

  It presumes it is run from a command-line application and supports the
  following flags:

    --auth_host_name: Host name to use when running a local web server
      to handle redirects during OAuth authorization.
      (default: 'localhost')

    --auth_host_port: Port to use when running a local web server to handle
    redirects during OAuth authorization.;
      repeat this option to specify a list of values
      (default: '[8080, 8090]')
      (an integer)

    --[no]auth_local_webserver: Run a local web server to handle redirects
      during OAuth authorization.
      (default: 'true')

  Since it uses flags make sure to initialize the gflags module before
  calling run().

  Args:
    flow: Flow, an OAuth 2.0 Flow to step through.
    storage: Storage, a Storage to store the credential in.
    http: An instance of httplib2.Http.request
         or something that acts like it.

  Returns:
    Credentials, the obtained credential.
  """
  logging.warning('This function, oauth2client.tools.run(), and the use of '
      'the gflags library are deprecated and will be removed in a future '
      'version of the library.')
  if FLAGS.auth_local_webserver:
    success = False
    port_number = 0
    for port in FLAGS.auth_host_port:
      port_number = port
      try:
        httpd = ClientRedirectServer((FLAGS.auth_host_name, port),
                                     ClientRedirectHandler)
      except socket.error as e:
        pass
      else:
        success = True
        break
    FLAGS.auth_local_webserver = success
    if not success:
      print('Failed to start a local webserver listening on either port 8080')
      print('or port 9090. Please check your firewall settings and locally')
      print('running programs that may be blocking or using those ports.')
      print()
      print('Falling back to --noauth_local_webserver and continuing with', end=' ')
      print('authorization.')
      print()

  if FLAGS.auth_local_webserver:
    oauth_callback = 'http://%s:%s/' % (FLAGS.auth_host_name, port_number)
  else:
    oauth_callback = client.OOB_CALLBACK_URN
  flow.redirect_uri = oauth_callback
  authorize_url = flow.step1_get_authorize_url()

  if FLAGS.auth_local_webserver:
    webbrowser.open(authorize_url, new=1, autoraise=True)
    print('Your browser has been opened to visit:')
    print()
    print('    ' + authorize_url)
    print()
    print('If your browser is on a different machine then exit and re-run')
    print('this application with the command-line parameter ')
    print()
    print('  --noauth_local_webserver')
    print()
  else:
    print('Go to the following link in your browser:')
    print()
    print('    ' + authorize_url)
    print()

  code = None
  if FLAGS.auth_local_webserver:
    httpd.handle_request()
    if 'error' in httpd.query_params:
      sys.exit('Authentication request was rejected.')
    if 'code' in httpd.query_params:
      code = httpd.query_params['code']
    else:
      print('Failed to find "code" in the query parameters of the redirect.')
      sys.exit('Try running with --noauth_local_webserver.')
  else:
    code = input('Enter verification code: ').strip()

  try:
    credential = flow.step2_exchange(code, http=http)
  except client.FlowExchangeError as e:
    sys.exit('Authentication has failed: %s' % e)

  storage.put(credential)
  credential.set_store(storage)
  print('Authentication successful.')

  return credential



import datetime
import time
import math
from decimal import Decimal
from django.db import models
from itertools import chain
from rest import helpers
from objict import objict
__MY_CACHE__ = objict()
# from . import profiler


def serialize(instance, graph):
    if isinstance(graph, str):
        graph = instance.getGraph(graph)
    return to_dict_from_graph(instance, graph)


def to_dict_from_graph(instance, graph):
    fields = graph.get("fields", [])
    extra = graph.get("extra", [])
    exclude = graph.get("exclude", [])
    recurse_into = graph.get("recurse_into", [])
    return to_dict(instance, fields=fields, extra=extra, exclude=exclude, recurse_into=recurse_into)


def get_from_cache(cache, instance):
    if cache is not None and isinstance(instance, models.Model) and hasattr(instance, "get_class_name"):
        cn = instance.get_class_name()
        key = f"{cn}.{instance.pk}"
        if key in cache:
            return cache[key]
    return None


def add_to_cache(cache, instance, data):
    if cache is not None and isinstance(instance, models.Model) and hasattr(instance, "get_class_name"):
        cn = instance.get_class_name()
        key = f"{cn}.{instance.pk}"
        if key not in cache:
            cache[key] = data


# @profiler.timeit
def to_dict(instance, fields=None, extra=[], exclude=[], recurse_into=[], cache=None):
    cached_item = get_from_cache(cache, instance)
    if cached_item:
        return cached_item
    # this is an attempt to work around the graphs at different depths not matching
    sub_cache = cache.get("sub_cache", None) if cache is not None else None
    if cache is not None and sub_cache is None:
        cache["sub_cache"] = dict()
        sub_cache = cache.get("sub_cache", None)

    fields = get_fields(instance) if not fields else fields
    data = {}
    recursed = set()
    combined_fields = list(chain(fields, extra, recurse_into))
    for f in combined_fields:
        fin, fout = (f[0], f[1]) if isinstance(f, (list, tuple)) else (f, f)
        if fin in exclude:
            continue

        split_fin = fin.split('.')
        if len(split_fin) > 1:
            top = split_fin[0]
            if top in recursed:
                continue
            recursed.add(top)
            fin = fout = top
        if fin.startswith("generic__") and hasattr(instance, "restGetGenericRelation"):
            fin = fin.split("__")[1]
            obj = get_generic_value(instance, fin, fout)
            if obj is not None:
                data[fin] = obj
                id_key = f"{fin}_id"
                if id_key in data:
                    del data[id_key]
            continue
        value = get_field_value(instance, fin)
        if isinstance(value, models.Model):
            if fin not in recursed:
                recursed.add(fin)
            if cache is not None:
                cached_item = get_from_cache(cache, value)
                if cached_item:
                    data[fout] = cached_item
                    continue
            sfields = get_model_fields(fin, combined_fields)
            if sfields:
                data[fout] = to_dict(value, sfields, cache=sub_cache)
            elif hasattr(value, "pk"):
                data[fout] = value.pk
                # FIXME we should use _id when not showing full model
                # data[f"{fout}_id"] = value.pk
            continue
        if isinstance(value, models.Manager):
            if fin not in recursed:
                recursed.add(fin)
            sfields = get_model_fields(fin, combined_fields)
            if __MY_CACHE__.cs is None:
                __MY_CACHE__.cs = helpers.importModule("rest.serializers.collection")
            sort = None
            qset = value.all()
            if hasattr(qset.model, "RestMeta") and hasattr(qset.model.RestMeta, "DEFAULT_SORT"):
                sort = qset.model.RestMeta.DEFAULT_SORT
            data[fout] = __MY_CACHE__.cs.to_list(qset, sort=sort, fields=sfields)["data"]
            continue

        if isinstance(value, dict) and len(split_fin) > 1:
            # let us do a complete reset here to get back to basics
            recursed.remove(fin)
            # now lets try and get the sub key
            top = split_fin.pop(0)
            value = helpers.getValueForKeys(value, split_fin)
            fin, fout = (f[0], f[1]) if isinstance(f, (list, tuple)) else (f, f)
            if "." in fout:
                keys = fout.split('.')
                tmp_data = data
                for k in keys[:-1]:
                    if k not in tmp_data:
                        tmp_data[k] = dict()
                        tmp_data = data[k]
                tmp_data[keys[-1]] = value
            else:
                data[fout] = value
            continue
        try:
            if callable(value):
                data[fout] = serialize_value(value())
            else:
                data[fout] = serialize_value(value)
        except Exception as err:
            data[fout] = str(err)
            helpers.log_exception(fout)
    if cache is not None:
        add_to_cache(cache, instance, data)
    return data


def serialize_value(value):
    if isinstance(value, (str, int, float)):
        return value
    if isinstance(value, (tuple, list)):
        value = [serialize_value(v) for v in value]
    elif isinstance(value, dict):
        value = {key: serialize_value(val) for key, val in value.items()}
    elif isinstance(value, datetime.datetime):
        # return int(time.mktime(value.timetuple()))
        value = value.timestamp()
    elif isinstance(value, datetime.date):
        value = value.strftime("%Y/%m/%d")
    elif isinstance(value, float) and math.isnan(value):
        value = 0.0
    elif isinstance(value, Decimal):
        value = float(value) if not value.is_nan() else 0.0
    return value


def get_generic_value(instance, key, graph="generic"):
    obj = None
    try:
        obj = instance.restGetGenericRelation(key)
    except Exception as err:
        print(err)
        return None
    if obj is None:
        return None
    if key == graph:
        graph = "generic"
    data = to_dict_from_graph(obj, obj.getGraph(graph))
    data["model"] = getattr(instance, key)
    return data


def get_field_value(instance, key):
    if isinstance(instance, dict):
        return instance.get(key, None)
    return getattr(instance, key) if hasattr(instance, key) else None


def get_model_fields(name, fields):
    skey = f"{name}."
    slen = len(skey)
    output = []
    for f in fields:
        if isinstance(f, (list, tuple)):
            if (f[0].startswith(skey)):
                output.append((f[0][slen:], f[1]))
        elif f.startswith(skey):
            output.append(f[slen:])
    return output
    # return [(f[0][f[0].find(".")+1:], f[1]) if isinstance(f, (list, tuple)) else f[f.find(".")+1:] for f in fields if f[0].startswith(skey)]


def get_fields(qset, model=None):
    fields = None
    if isinstance(qset, dict):
        return list(qset.keys())
    elif hasattr(qset, '_meta'):
        fields = qset._meta.fields
    elif hasattr(qset, 'model'):
        fields = qset.model._meta.fields
    elif hasattr(qset, 'keys'):
        return list(qset.keys())
    elif model and hasattr(model, '_meta'):
        fields = model._meta.fields
    return [f.name for f in fields] if fields else None


def expand_fields(fields):
    output = dict()
    for key in fields:
        if "." in key:
            fields = key.split('.')
            skey = fields.pop(0)
            if output.get(skey, None) is None:
                output[skey] = []
            output[skey].append(".".join(fields))
        elif key not in output:
            output[key] = None
    expanded = objict()
    for key, value in output.items():
        if value is not None:
            expanded[key] = expand_fields(value)
        else:
            expanded[key] = None
    return expanded


def get_foreign_fields(model):
    return [field.name for field in model._meta.fields if field.get_internal_type() == "ForeignKey"]


def get_select_related_fields(model, fields):
    if isinstance(fields, list):
        fields = expand_fields(fields)
    fk_fields = get_foreign_fields(model)

    related_fields = []
    for key, value in fields.items():
        if value is not None and key in fk_fields:
            related_fields.append(key)
            fmodel = model.get_fk_model(key)
            for related in get_select_related_fields(fmodel, value):
                related_fields.append(f"{key}__{related}")
    return related_fields            
